import re
import datetime
import warnings
from typing import Optional, List, Union, Callable, Iterable, cast
from collections import Counter

import os

from mhelper import array_helper


__strip_lines_rx = re.compile( r"^[ ]+", re.MULTILINE )


def first_words( t: str, min_length = 1 ) -> str:
    """
    Returns a string made up of the first and last words (optionally allows a min_length of first words)
    """
    result = ""
    last_iter = ""
    for match in re.finditer( r"[\w]+", t ):
        if len( match.group( 0 ) ) <= 1:
            continue
        
        last_iter = match.group( 0 )
        
        if len( result ) < min_length:
            if result:
                result += " "
            result += last_iter
            last_iter = None
    if result:
        if last_iter:
            return result + " " + last_iter
        else:
            return result
    else:
        return t


def first_word( t: str ) -> str:
    """
    Returns the first word from the string
    """
    match = re.match( r"[\w]+", t )
    
    if not match:
        return t
    
    return match.group( 0 )


def time_to_string( time: float ) -> str:
    """
    Given a time in seconds, returns an approximate string representation like "5 seconds", "4 minutes", etc.
    """
    if time <= 1.5:
        return str( round( time * 1000 ) ) + " milliseconds"
    
    if time < 60:
        return str( round( time ) ) + " seconds"
    
    time /= 60
    
    if time < (60 * 2):
        return str( round( time ) ) + " minutes"
    
    time /= 60
    
    return str( round( time ) ) + " hours"


def time_to_string_short( time: float, delimiter: str = ":" ) -> str:
    """
    Given a time in seconds, returns a formatted string like "00:05", or "02:01:05".
    """
    SECONDS_IN_ONE_HOUR = 60 * 60
    SECONDS_IN_ONE_MINUTE = 60
    
    hours = time // SECONDS_IN_ONE_HOUR
    time -= hours * SECONDS_IN_ONE_HOUR
    
    minutes = time // SECONDS_IN_ONE_MINUTE
    time -= minutes * SECONDS_IN_ONE_MINUTE
    
    seconds = time
    
    h_text = ""
    
    if hours:
        h_text = str( int( hours ) ) + delimiter
    
    m_text = str( int( minutes ) ).rjust( 2, "0" ) + delimiter
    
    s_text = str( int( seconds ) ).rjust( 2, "0" )
    
    return h_text + m_text + s_text


def highlight_words( text: str, words, colour, normal ):
    text = normal + text
    
    for x in words:
        text = highlight_regex( text, "/\b($" + x + ")\b/i", colour, normal )
    
    return text


def highlight_regex( text, regex, colour, normal, group = 1 ):
    l = len( colour ) + len( normal )
    
    for i, m in enumerate( re.finditer( regex, text, re.IGNORECASE ) ):
        s = m.start( group ) + l * i
        e = m.end( group ) + l * i
        
        text = text[:s] + colour + text[s:e] + normal + text[e:]
    
    return text


def highlight_quotes( text, start, end, colour, normal ):
    find = start + "([^" + start + end + "]+)" + end
    replace = colour + "\\1" + normal
    return re.sub( find, replace, text )


def curtail( text: str, start: Optional[str] = None, end: Optional[str] = None, error = False ):
    """
    Removes text from the start of end of a string
    """
    if start:
        if text.startswith( start ):
            text = text[len( start ):]
        elif error:
            raise KeyError( "Trying to remove the substring «{0}» from the string «{1}» but the string does not start with the substring.".format( start, text ) )
    
    if end:
        if text.endswith( end ):
            text = text[:len( text ) - len( end )]
        elif error:
            raise KeyError( "Trying to remove the substring «{0}» from the string «{1}» but the string does not end with the substring.".format( end, text ) )
    
    return text


def time_now() -> str:
    t = datetime.datetime.now()
    return t.strftime( '%Y-%m-%d %H:%M:%S' )


def percent( val ):
    val *= 100
    return "{0:.1f}%".format( val )


def timedelta_to_string( delta ) -> str:
    s = delta.seconds
    hours = s // 3600
    s -= hours * 3600
    minutes = s // 60
    seconds = s - (minutes * 60)
    
    if hours:
        return '%d:%02d:%02d' % (hours, minutes, seconds)
    else:
        return '%02d:%02d' % (minutes, seconds)


def fix_width( text: object, width: int = 20 ) -> str:
    return max_width( str( text ), width ).ljust( width )


def max_width( text: str, width: int = 20 ) -> str:
    if not isinstance( text, str ):
        warnings.warn( "max_width expects a `str` but got a «{}»".format( type( text ) ) )
        text = str( text )
    
    if width <= 1:
        warnings.warn( "Did you really mean to pass <width = {}> to max_width?".format( width ) )
    
    text = text.strip()
    
    if "\n" in text:
        text = text[:text.index( "\n" )]
    
    if len( text ) > width:
        text = text[:(width - 1)] + "…"
    
    assert len( text ) <= width
    return text


def size_to_string( size: int ):
    """
    Returns a `size`, specified in bytes, as a human-readable string, similar to `ls -h` in bash.
    """
    if size == -1:
        return "?"
    
    if size < 1024:
        return "{0:.1f}b".format( size )
    
    size /= 1024
    
    if size < 1024:
        return "{0:.1f}kb".format( size )
    
    size /= 1024
    
    if size < 1024:
        return "{0:.1f}Mb".format( size )
    
    size /= 1024
    
    return "{0:.1f}Gb".format( size )


def summarised_join( source: Counter, delimiter ):
    r = []
    
    for k, v in sorted( source.items(), key = lambda x: -x[1] ):
        r.append( "{1} × {0}".format( k, v ) )
    
    return delimiter.join( r )


def is_int( x ):
    if not x:
        return False
    
    try:
        _ = int( x )
        return True
    except:
        return False


def bulk_replace( text, format = "<*>", **kwargs ):
    for k, v in kwargs.items():
        text = text.replace( format.replace( "*", k ), v )
    
    return text


__make_name_regex = re.compile( "[^0-9a-zA-Z.]" )


def make_name( name ):
    return __make_name_regex.sub( "_", name )


def strip_lines( text ):
    return __strip_lines_rx.sub( "", text )


def prefix_lines( text, prefix, suffix = "" ):
    return prefix + text.replace( "\n", suffix + "\n" + prefix ) + suffix


def remove_prefix( text, prefix ) -> str:
    if not text.startswith( prefix ):
        raise ValueError( "I've been asked to remove a prefix «{0}» from a string «{1}», but the string doesn't start with that prefix.".format( prefix, text ) )
    
    return text[len( prefix ):]


def type_name( value ):
    return type( value ).__name__


def undo_camel_case( text: str, sep = " " ):
    result = []
    
    for i, c in enumerate( text ):
        if i and c.isupper():
            result.append( sep )
        
        result.append( c )
    
    return "".join( result )


def name_index( names: List[str], name: str ) -> int:
    """
    Given a `name`, finds its index in `names`. `name` can be a `str`, an `int` or an int as a str.
    :param names: List of names 
    :param name:  Name or index to find 
    :return: Index
    :exception ValueError: Not found
    """
    if isinstance( name, int ):
        v = name
    elif not isinstance( name, str ):
        raise TypeError( "`name` should be an `int` or a `str`, but it is «{0}» which is a {1}.".format( name, type_name( name ) ) )
    elif all( str.isdigit( x ) for x in name ):
        v = int( name )
    else:
        v = None
    
    if v is not None and 0 <= v < len( names ):
        return v
    
    if name in names:
        return names.index( name )
    
    raise ValueError( "Trying to find the column with header «{0}» but that column doesn't exist or is out of range. The columns are: {1}".format( name, names ) )


def list_to_string( list, sep = ", ", format = "{0}" ):
    return format.format( sep.join( str( x ) for x in list ) )


def current_time():
    import time
    return time.strftime( "%Y-%m-%d %H:%M:%S %z" )


def ordinal( i: object ):
    """
    Returns "1st", "2nd", "3rd", etc. from 1, 2, 3, etc.
    `i` can be anything coercible to a string that looks like a number.
    """
    i = str( i )
    l = i[-1]
    
    if l == "1":
        return i + "st"
    elif l == "2":
        return i + "nd"
    elif l == "3":
        return i + "rd"
    else:
        return i + "th"


def get_indent( line ):
    num_spaces = 0
    for x in line:
        if x == " ":
            num_spaces += 1
        else:
            return num_spaces
    
    return num_spaces


def remove_indent( current_indent, line: str ):
    i = 0
    
    for i, x in enumerate( line ):
        if x != " ":
            break
    
    i = min( i, current_indent )
    
    return line[i:]


def capitalise_first_and_fix( text: Optional[object], swap = "_-" ) -> str:
    text = capitalise_first( text )
    
    for x in swap:
        text = text.replace( x, " " )
    
    return text


def capitalise_first( text: Optional[object] ) -> str:
    if text is None:
        return ""
    
    text = str( text )
    
    if not text:
        return text
    
    if len( text ) == 1:
        return text[0].upper()
    
    return text[0].upper() + text[1:]


def special_to_symbol( value ):
    return value.replace( "\n", "␊" ).replace( "\r", "␍" ).replace( "\t", "␉" )


def unescape( v ):
    try:
        return v.encode().decode( "unicode_escape" )
    except Exception as ex:
        raise ValueError( "Unescape «{}» failed.".format( v ) ) from ex


def to_bool( x ):
    try:
        y = float( x )
        return y != 0
    except:
        x = x.lower()
        if x in ("true", "yes", "t", "y", "1", "on"):
            return True
        elif x in ("false", "no", "f", "n", "0", "off"):
            return False
        else:
            raise ValueError( "Cannot convert «{}» to boolean.".format( x ) )


DPrint = Callable[[object], None]


class FindError( Exception ):
    pass


def find( *,
          source: Iterable[object],
          search: str,
          namer: Optional[Union[Callable[[object], str], Callable[[object], List[str]]]] = None,
          detail: Optional[str] = None,
          fuzzy: Optional[bool] = True ):
    """
    Finds the command or plugin with the closest name to "text".
    
    :param source:  Source list 
    :param search:  Text to find 
    :param namer:   How to translate items in the source list, each item can translate to a `str`, or a `list` or `tuple` of `str` 
    :param detail:  What to call the collection in error messages.
    :param fuzzy:   Permit partial (start-of-word) matches
    :return:        The matching item in `source`
    :except FindError: The text was not matched 
    """
    
    #
    # Arguments
    #
    if not isinstance( source, list ) and not isinstance( source, tuple ):
        source = list( source )
    
    if not isinstance( search, str ):
        raise TypeError( "string_helper.find() takes a `str` and not a {} (`{}`).".format( type( search ).__name__, search ) )
    
    if namer is None:
        namer = cast( Callable[[object], str], str )
    
    
    def __get_names( item ):
        r = namer( item )
        
        if isinstance( r, str ):
            return [r]
        elif isinstance( r, list ) or isinstance( r, tuple ):
            return r
        else:
            raise TypeError( "Return value of namer «{}».".format( namer ) )
    
    
    #
    # Exact match
    #
    for item in source:
        for name in __get_names( item ):
            if search == name:
                return item
    
    #
    # Start-of-word match
    #
    if fuzzy:
        match_items = set()
        match_names = set()
        
        for item in source:
            for name in __get_names( item ):
                if name.startswith( search ):
                    match_items.add( item )
                    match_names.add( name )
        
        if len( match_items ) == 1:
            return array_helper.first( match_items )
    else:
        match_items = set()
    
    #
    # Failure
    #
    if detail is None:
        detail = "item"
    
    if not match_items:
        ss = []
        ss.append( "No such {}: «{}»".format( detail, search ) )
        available = []
        
        for item in source:
            for name in __get_names( item ):
                if name:
                    available.append( name )
        
        ss.append( "Options: " + ", ".join( "«{}»".format( x ) for x in available ) )
        
        raise FindError( "\n".join( ss ) )
    
    ss = []
    ss.append( "Ambiguous {} name: «{}»".format( detail, search ) )
    ss.append( "    ...did you mean..." )
    for item in match_items:
        for name in __get_names( item ):
            ss.append( "    ...... " + name.replace( search, "«" + search + "»" ) )
    
    raise FindError( "\n".join( ss ) )


def no_emoji( x ):
    return x + "\uFE0E"


def fix_indents( text: str ) -> str:
    text = str( text )
    lines = text.split( "\n" )
    lines[0] = lines[0].lstrip()
    
    if len( lines ) == 1:
        return lines[0]
    
    min_leading = 9999
    
    for line in lines[1:]:
        stripped = line.lstrip()
        
        if stripped:
            leading = len( line ) - len( stripped )
            min_leading = min( leading, min_leading )
    
    if min_leading == 9999:
        return text.strip()
    
    for i in range( 1, len( lines ) ):
        lines[i] = lines[i][min_leading:]
    
    return "\n".join( lines )


def assert_unicode():
    """
    UTF-8 check. Probably Windows Console Services is badly configured.
    """
    UNICODE_ERROR_MESSAGE = \
        """


        +-------------------------------------------------------------------------------------------------------+--------+
        | UNICODE ENCODE ERROR                                                                                  |        |
        +-------------------------------------------------------------------------------------------------------+  X  X  +
        | It looks like your console doesn't support Unicode.                                                   |        |
        |                                                                                                       |  _--_  |
        | intermake needs Unicode to display its UI.                                                             |        |
        | Python doesn't seem to know your terminal doesn't support UTF8 and just crashes :(                    +--------|
        | This problem usually occurs when using cmd.exe on Windows.                                                     |
        |                                                                                                                |
        | On Windows, you could `set PYTHONIOENCODING=ascii:replace` as a quick fix, but it might be better in the long  |
        | run to setup your console and font to support UTF8: https://stackoverflow.com/questions/379240                 |
        |                                                                                                                |
        | On Unix, the quick fix is `export PYTHONIOENCODING=ascii:replace`                                              |
        +----------------------------------------------------------------------------------------------------------------+


        """
    try:
        print( "😁\r \r", end = "" )
    except UnicodeEncodeError as ex:
        # On the plus side, Window's Console has a horizontal scroll bar, so we can display an over-sized error message without it breaking...
        raise ValueError( UNICODE_ERROR_MESSAGE ) from ex


def centre_align( text, width, char = " ", prefix = None, suffix = None ):
    """
    Centre aligns text.
    :param suffix:  Prefix to use in output. e.g. colour codes that shouldn't be considered as part of the length.
    :param prefix:  Suffix to use in output. e.g. colour codes that shouldn't be considered as part of the length.
    :param text:    Text to align 
    :param width:   Width to align into 
    :param char:    Padding character
    :return:        Aligned text. 
    """
    
    use = text
    
    if prefix:
        use = prefix + use
    
    if suffix:
        use += suffix
    
    length = len( text )
    
    pad_len = (width - length) // 2 - 1
    pad = char * pad_len
    
    if (length % 2) != 0 and (length + pad_len * 2) < width:
        return pad + use + pad + char
    else:
        return pad + use + pad



from typing import Any, Callable, Optional, Type, TypeVar, Union, cast, Sequence, Generic

# Decorator function, returns a function that takes and returns a function
from mhelper import array_helper


DECORATOR_FUNCTION = Callable[..., Callable[[Callable], Callable]]
# Decorator - takes a function, returns a function
DECORATOR = Callable[[Callable], Callable]
DECORATED = Optional[object]

T = TypeVar( "T" )


class MGenericMeta( type ):
    """
    Base class for generics.
    For example see `MGeneric` documentation.
    
    :data __parameters__:   Field containing the generic parameters.
                            Empty for the non-generic base.
                            
    :remarks: Nb. Name of `__parameters__` violates PEP8 but is coherent with PEP464 implementation and signifies a meta-field. 
    """
    
    
    def __new__( mcs, name, bases, namespace, parameters = None ):
        cls = super().__new__( mcs, name, bases, namespace )
        return cls
    
    
    def __init__( cls, name, bases, namespace, parameters = None ):
        super().__init__( name, bases, namespace )
        
        if parameters is None:
            if bases:
                for base in bases:
                    if hasattr( base, "__parameters__" ):
                        if parameters is not None:
                            raise TypeError( "Cannot inherit parameters because I don't know which subclass to use." )
                        
                        parameters = base.__parameters__
            
            if parameters is None:
                parameters = tuple()
        
        cls.__parameters__ = parameters
        cls.__CACHE = { }
    
    
    def __getitem__( cls: T, parameters ) -> T:
        """
        Obtains a class with the specified generic parameters.
        Results are cached, so asking for the same parameters again yields the same class instance.
        
        If the called class itself has generic parameters, the parameters are concatenated, for instance all of:
            A = MGeneric[int, bool]
            B = MGeneric[int][bool]
            C = MGeneric[int][bool]
        Have the same generic parameters (int, bool).
        However, the instances are different, hence:
            A.__parameters__ == B.__parameters__ == C.__parameters__
            A is not B
            B is C 
        
        :param parameters: Generic parameters. 
        """
        if parameters is None:
            parameters = ()
        elif not isinstance( parameters, tuple ):
            parameters = (parameters,)
        
        if parameters in cls.__CACHE:
            return cls.__CACHE[parameters]
        
        if cls.__parameters__:
            all_parameters = cls.__parameters__ + parameters
        else:
            all_parameters = parameters
        
        result = cls.__class__( cls.__name__,
                                (cls,) + cls.__bases__,
                                dict( cls.__dict__ ),
                                parameters = all_parameters )
        
        cls.__CACHE[parameters] = result
        
        return result
    
    
    def __str__( self ):
        """
        Returns a string similar to normal Python classes, but includes the generic parameters.
        Note that this is not the same as `__name__`, which always holds only the base class name, without the generics.
        """
        if self.__parameters__:
            return "<class {}{}>".format( self.__name__, ("[" + (", ".join( str( x ) for x in self.__parameters__ ) + "]")) )
        else:
            return "<class {}>".format( self.__name__ )


class MGeneric( metaclass = MGenericMeta ):
    """
    Instantiation of class with MGenericMeta meta-class.
    
    EXAMPLE:
        
        ```
        class MyList( MGeneric ):
            def item_type(self):
                return self.__parameters__[0]
            
            def append( item ):
                assert isinstance(item, self.item_type())
                ...
        ```
    """
    pass


# noinspection PyUnusedLocal
def placeholder( type_: Type[T] ) -> T:
    """
    Always returns `None`.
    For fixing lint errors where the specified type is expected
    """
    # noinspection PyTypeChecker
    return None


class MAnnotationFactory:
    """
    Class that generates objects intended to be used as a function parameter annotations.
    See `MAnnotation` for usage.
    """
    
    
    def __init__( self, name: str,
                  processing: "Optional[Callable[[MAnnotation], None]]" = None,
                  annotation_type: "Type[MAnnotation]" = None ):
        """
        CONSTRUCTOR
        :param name:            Name of the annotation class (generally the same as the variable name). 
        :param processing:      If not none, specifies a function called on all new annotations.
        :param annotation_type: If not none, specifies the type of MAnnotation to create. 
        """
        self.name = name
        self.processing = processing or None
        self.children = { }
        self.annotation_type = MAnnotation if annotation_type is None else annotation_type
    
    
    def __str__( self ) -> str:
        """
        Name of the factory.
        """
        return self.name
    
    
    def __getitem__( self, item ) -> "MAnnotation":
        """
        Obtains an annotation.
        :param item:    Interior annotation
        """
        return self.create( item, None )
    
    
    def create( self, child: object, parameters: Optional[Sequence[object]] ) -> "MAnnotation":
        """
        Obtains an annotation.
        
        :param child:        Interior annotation, usually the type or another annotation, but can be anything
        :param parameters:   Special parameters on the annotation to retrieve
        :return: The annotation
        """
        parameters = array_helper.as_sequence( parameters, cast = list )
        
        key = tuple([child] + list( parameters ))
        
        try:
            result = self.children.get( key )
        except TypeError as ex:
            raise TypeError("Invalid annotation parameters: «{}» on «{}».".format(", ".join(str(x) for x in key), self)) from ex
        
        if result is None:
            result = self.annotation_type( MAnnotationArgs( self, child, parameters ) )
            self.children[key] = result
        
        return result
    
    
    def __contains__( self, item ) -> bool:
        """
        Returns if the specified annotation is a member of this factory.
        """
        return isinstance( item, MAnnotation ) and item.parent is self


class MAnnotationArgs:
    def __init__( self,
                  factory: Union[MAnnotationFactory, str] = None,
                  child: object = None,
                  parameters: Optional[Sequence[object]] = None ):
        """
        :param factory:     Factory this annotation was created by 
        :param child:       Sub-type of the annotation, generally a type of an MAnnotation, but can be anything. 
        :param parameters:  Parameters on this annotation. 
        """
        self.factory = factory
        self.child = child
        self.parameters = parameters


class MAnnotation:
    """
    Class for objects intended to be used as a function parameter annotations.
    
    ```
    FileName = MAnnotationFactory("FileName")[ str ]
    
    def print_file(file_name : FileName):
        print(open(file_name).read())
    ```
    
    ```
    NonZero = MAnnotationFactory("NonZero")
    
    def divide_int(a : int, b : NonZero[int]):
        return a // b
    ```
    
    ```
    def fn(a):
        a.extension = a.parameters[0]
    
    FileName = MAnnotationFactory("FileName", processing = fn)[ str ]
    
    def print_file(file_name : FileName[".txt"]):
        print(open(file_name).read())
        
    def main():
        f = input()
        
        if not f.endswith(print_file.__annotations__["file_name"].extension):
            return 1
            
        print_file(f)
    ```
    """
    
    
    def __init__( self, args: MAnnotationArgs ):
        """
        CONSTRUCTOR
        :param args:      Constructor arguments 
        """
        parameters = array_helper.as_sequence( args.parameters )
        
        self.factory = args.factory
        self.child = args.child
        self.parameters = parameters
        
        if not isinstance(args.child, type) and not isinstance(args.child, MAnnotation) and not isinstance(args.child, type(Union[int,str])) and not isinstance(args.child, tuple):
            raise TypeError("args.child should be a type or another MAnnotation, but it's a «{}» («{}»).".format(type(args.child), args.child))
        
        if self.factory.processing is not None:
            self.factory.processing( self )
    
    def __str__( self ) -> str:
        return "{}[{}]".format( self.factory, self.child, ", {}".format( self.parameters) if self.parameters else "" )
    
    
    def __getitem__( self, item ) -> "MAnnotation":
        """
        Obtains a further annotation with the specified parameters.
        The parameters must have been specified on the `parent`.
        
        :param item:    Parameter sequence 
        """
        if self.parameters:
            raise ValueError("This is already the parameterised instance, «{}», cannot parameterise again using «{}».".format(self, item))
        
        return self.factory.create( self.child, item )


class GenericStringMeta( MGenericMeta ):
    def __subclasscheck__( self, cls ):
        if cls is Any:
            return True
        if self.__parameters__ is None:
            return isinstance( cls, GenericStringMeta )
        elif isinstance( cls, GenericStringMeta ):
            return True
        elif isinstance( cls, GenericStringMeta ):
            if issubclass( cls, str ):
                return True
            return False
        else:
            return issubclass( cls, str )
    
    
    def __instancecheck__( self, obj ):
        print( "__instancecheck__ {}".format( obj ) )
        return self.__subclasscheck__( type( obj ) )


class GenericString( str, metaclass = GenericStringMeta ):
    """
    Something that is a string.
    
    This property acts as a hint to the UI, indicating the string's values are constrained.
    
    A "label" of the node or edge can be specified as a generic argument.
    Use "@" to reference the name of another function argument.
    
    
    GenericString["label"]("value")
        - OR -
    GenericString("value")
    """
    
    
    def __new__( cls: Type[T], *args, **kwargs ) -> T:
        # noinspection PyArgumentList,PyTypeChecker
        return cast( cls, str.__new__( cls, *args, **kwargs ) )
    
    
    # noinspection PyUnresolvedReferences
    @classmethod
    def type_label( cls ):
        if cls.__parameters__:
            return cls.__parameters__[0]
        else:
            return None


class NonGenericString( str ):
    """
    NonGenericString("value")
    
    This property acts as a hint to the UI, indicating the string's values are constrained.
    """
    
    
    def __new__( cls: Type[T], *args, **kwargs ) -> T:
        # noinspection PyArgumentList,PyTypeChecker
        return cast( cls, str.__new__( cls, *args, **kwargs ) )


class TypedList( MGeneric ):
    def get_list_type( self ):
        return self.__parameters__[0]
    
    
    def __init__( self, *args, **kwargs ):
        self.__list = list( *args, **kwargs )
    
    
    def __getitem__( self, item ):
        return self.__list[item]
    
    
    def __setitem__( self, key, value ):
        self.__check_instance( value )
        self.__list[key] = value
    
    
    def append( self, value ):
        self.__check_instance( value )
        self.__list.append( value )
    
    
    def __check_instance( self, value ):
        if not isinstance( value, self.__parameters__[0] ):
            raise TypeError( "Value «{}» of incorrect type «{}» added to list of «{}».".format( value, type( value ), self.get_list_type() ) )


class ByRef( Generic[T] ):
    """
    Pass value by reference.
    """
    
    
    def __init__( self, value: T ):
        self.value: T = value



from typing import Union


class Logger:
    def __init__( self, name: Union[ str, bool ], enabled: bool = None ):
        if isinstance( name, bool ) and enabled is None:
            enabled = name
            name = "untitled"
        
        if name:
            self.__prefix = name + ": "
        else:
            self.__prefix = ""
        
        if not enabled:
            self.__print = self.__print_nowhere
        else:
            self.__print = print
    
    
    def __print_nowhere( self, *args, **kwargs ):
        pass
    
    
    def __call__( self, *args, **kwargs ):
        if len( args ) == 1:
            self.print( args[ 0 ] )
        elif len( args ) > 1:
            self.print( args[ 0 ].format( *args[ 1: ] ) )
        
        return self
    
    
    def print( self, message ):
        self.__print( self.__prefix + message )
    
    
    def indent( self ):
        self.__prefix += "    "
        
        if len( self.__prefix ) > 100:
            from mhelper.exception_helper import ImplementationError
            raise ImplementationError( "Log indent too far." )
    
    
    def unindent( self ):
        self.__prefix = self.__prefix[ :len( self.__prefix ) - 4 ]
    
    
    def __enter__( self ):
        self.indent()
    
    
    def __exit__( self, exc_type, exc_val, exc_tb ):
        self.unindent()



"""
Helper functions for arrays, lists, iterable, etc.
"""

import warnings
from typing import List, Optional, Iterator, Tuple, Dict, Iterable, Union, TypeVar, Callable, cast


T = TypeVar( "T" )
U = TypeVar( "U" )


def list_type( the_list ) -> type:
    """
    Determines the type of elements in a list

    Errors if the list doesn't contain any elements, or if the elements are of varying type

    :param the_list: List to check
    :return: Type of elements in the list
    """
    
    t = None
    
    for e in the_list:
        et = type( e )
        
        if t is None:
            t = et
        elif t is not et:
            raise ValueError( "Calling list_type on a list with at least two types ({0} and {1})".format( t, et ) )
    
    if t is None:
        raise ValueError( "Calling list_type on a list with no elements." )
    
    return t


def dict_to_list( v: dict ) -> List[str]:
    """
    Converts a string dictionary to a string list where each element is "x = y"
    """
    
    r = []
    
    for k, v in v.items():
        r.append( "{}={}".format( k, v ) )
    
    return r


def decomplex( item: Union[Optional[object], List[Optional[object]]] ) -> Optional[object]:
    """
    Converts items in arrays to single items
    """
    
    # noinspection PyBroadException
    try:
        if len( item ) == 1 and item[0] is not item:
            return decomplex( item[0] )
    except:
        pass
    
    return item


def as_sequence( contents: Union[T, List[T], Tuple[T]], elemental_none = False, cast = None, accept = (list, tuple) ) -> Union[List[T], Tuple[T]]:
    """
    The contents, as a tuple or list, if they are a tuple or list
    Or the sole item in a tuple
    
    Note: only `tuple` and `list` are handled, other `iterable`s are still placed in a tuple because we cannot know their nature.
    
    :param contents:        List contents
    :param elemental_none:  Doesn't treat `None` as an empty list
    :param cast:            Casts the result to this list type, if it isn't already 
    :param accept:          Types to accept
    :return: The result, as a list
    """
    if not elemental_none and contents is None:
        contents = tuple()
    elif not any( isinstance( contents, x ) for x in accept ):
        contents = contents,
    
    if cast is not None and not isinstance( contents, cast ):
        contents = cast( contents )
    
    return contents


def as_iter( contents: Union[T, List[T], Tuple[T]] ) -> Tuple[T]:
    """
    DEPRECATED
    """
    warnings.warn( "as_iter is deprecated, use `as_sequence`.", DeprecationWarning )
    return as_sequence( contents, elemental_none = True )


def as_tuple( contents: Union[T, List[T], Tuple[T]] ) -> Tuple[T]:
    """
    DEPRECATED
    """
    warnings.warn( "as_tuple is deprecated, use `as_sequence`.", DeprecationWarning )
    return as_sequence( contents, elemental_none = True, cast = tuple )


def as_list( contents: Union[T, List[T], Tuple[T]] ) -> List[T]:
    """
    DEPRECATED
    """
    warnings.warn( "as_list is deprecated, use `as_iter`.", DeprecationWarning )
    return as_sequence( contents, elemental_none = True, cast = list )


def backwards_range( count ):
    """
    A range, going backwards.
    """
    return range( count - 1, -1, -1 )


# !has test case
def create_index_lookup( source: Iterable[T] ) -> Dict[T, int]:
    """
    Creates a lookup table (`dict`) that allows the index of elements in `the_list` to be found.
    """
    result = { }
    result.update( (v, i) for i, v in enumerate( source ) )
    
    return result


def deinterleave_as_two( source: Iterable[T] ) -> Tuple[List[T], List[T]]:
    """
    Deinterleaves a source list, returns two new lists
    """
    dest_a = []
    dest_b = []
    iterator = iter( source )
    
    for a in iterator:
        dest_a.append( a )
        dest_b.append( next( iterator ) )
    
    return dest_a, dest_b


def deinterleave_as_iterator( source: Iterable[T] ) -> Iterator[Tuple[T, T]]:
    """
    Deinterleaves a source list, returns an iterator over tuples
    """
    iterator = iter( source )
    
    for a in iterator:
        yield a, next( iterator )


def deinterleave_as_list( source: Iterable[T] ) -> List[Tuple[T, T]]:
    """
    Deinterleaves a source list "A,B,A,B,...", returns a list of tuples "A, B"
    """
    dest_a = []
    iterator = iter( source )
    
    for a in iterator:
        dest_a.append( (a, next( iterator )) )
    
    return dest_a


def deinterleave_as_dict( source ) -> Dict[object, Optional[object]]:
    """
    Deinterleaves a source list "K,V,K,V,...", returns a dictionary "D" of "V = D[K]"
    """
    return dict( deinterleave_as_iterator( source ) )


def has_any( sequence: Iterable ) -> bool:
    """
    Returns if the sequence contains _any_ elements (i.e. non-zero length).
    """
    for _ in sequence:
        return True
    
    return False


def iterate_all( root, fn = None ) -> Iterator[Optional[object]]:
    """
    Iterates all items and descendants
    :param root: Where to start 
    :param fn: How to get the children 
    :return: Iterator over items and all descendants 
    """
    if fn is None:
        fn = lambda x: x
    
    for x in fn( root ):
        yield x
        yield from iterate_all( x, fn )


def ensure_capacity( array, index, value = None ):
    if len( array ) <= index:
        needed = index + 1 - len( array )
        array += [value] * needed


def index_of_first( array, elements ) -> Optional[int]:
    for e in elements:
        for i, x in enumerate( array ):
            if x == e:
                return i
    
    return None


class Indexer:
    """
    Provides a name to index and index to name lookup table.
    
    Note that `Indexer` has no indexer - be specific, use:
        `Indexer.name`
        `Indexer.index`
    """
    
    
    def __init__( self, iterator: Iterable[object] = None ):
        """
        CONSTRUCTOR
        Allows initialisation from existing entries
        """
        self.indexes = { }  # names to indices
        self.names = []  # indices to names
        
        if iterator is not None:
            for name in iterator:
                self.add( name )
    
    
    def add( self, name: object ):
        """
        Adds a new name with a new index.
        """
        index = self.indexes.get( name )
        
        if index is not None:
            return index
        
        index = len( self )
        
        self.indexes[name] = index
        self.names.append( name )
        
        return index
    
    
    def __len__( self ) -> int:
        """
        OVERRIDE
        Obtains the number of names
        """
        return len( self.names )
    
    
    def index( self, name: object ) -> int:
        """
        Obtains the index of the specified name.
        """
        return self.indexes[name]
    
    
    def name( self, index: int ):
        """
        Obtains the name at the specified index.
        """
        return self.names[index]


def first( array: Iterable[T] ) -> Optional[T]:
    for item in array:
        return item
    
    return None


def first_unsafe( array: Iterable[T] ) -> T:
    """
    Returns the first element of the array, raising a `KeyError` on failure.
    """
    for item in array:
        return item
    
    raise KeyError( "Cannot extract the first element of the iterable because the iterable has no elements." )


def extract_single_item( array: Iterable[T] ) -> Optional[T]:
    """
    Returns the first element of the array, returning `None` if the length is not `1`.
    """
    it = iter( array )
    
    try:
        first = next( it )
    except StopIteration:
        return None
    
    try:
        next( it )
        return None
    except StopIteration:
        return first


def lagged_iterate( sequence: Iterable[Optional[T]] ) -> Iterator[Tuple[Optional[T], Optional[T]]]:
    """
    Iterates over all adjacent pairs in the sequence, (0,1),(1,2),(2,3),(...,...),(n-1,n)
    """
    has_any = False
    previous = None
    
    for current in sequence:
        if has_any:
            yield previous, current
        
        has_any = True
        previous = current


def triangular_comparison( sequence: List[Optional[T]] ) -> Iterator[Tuple[Optional[T], Optional[T]]]:
    for index, alpha in enumerate( sequence ):
        for beta in sequence[index + 1:]:
            yield alpha, beta


class Single:
    def __init__( self, value = 0 ):
        self.value = value
    
    
    def next( self ):
        self.value += 1
        return self.value


def sort_two( a: T, b: U ) -> Tuple[T, U]:
    if a > b:
        return b, a
    else:
        return a, b


def sort_two_by_key( a: T, b: U, ka, kb ) -> Tuple[T, U]:
    if ka > kb:
        return b, a
    else:
        return a, b


def ordered_insert( list: List[T], item: T, key: Callable[[T], object] ):
    """
    Inserts the `item` into the `list` that has been sorted by `key`.
    """
    import bisect
    list.insert( bisect.bisect_left( [key( x ) for x in list], key( item ) ), item )


def average( list ) -> float:
    """
    Returns the mean average.
    """
    return sum( list ) / len( list )


def count( list ):
    try:
        return len( list )
    except:
        return sum( 1 for _ in list )


def when_last( i ):
    f = True
    l = None
    
    for x in i:
        if not f:
            yield l, False
        else:
            f = False
        
        l = x
    
    yield l, True


def make_dict( **kwargs ):
    return kwargs





class Event:
    def __init__(self):
        self.targets = set()
        
    def __iadd__(self, other):
        self.targets.add(other)
        
    def __isub__(self, other):
        self.targets.remove(other)
        
    def __call__(self, *args, **kwargs):
        for target in self.targets:
            target(*args, **kwargs)
        
        


from typing import Tuple


def increment_mean(mean : float, count : int, new_value : float) -> Tuple[float, int]:
    new_count = count + 1 
    new_mean = mean + ((new_value - mean) / new_count)
    return new_mean, new_count 


from PyQt5.QtCore import pyqtSlot
from PyQt5.QtWidgets import QDialog, QDialogButtonBox

from mhelper.designer.frm_text_designer import Ui_Dialog


class FrmText(QDialog):
    def __init__( self, parent, message, text ):
        """
        CONSTRUCTOR
        """
        QDialog.__init__( self, parent )
        self.ui = Ui_Dialog( self )
        self.setWindowTitle(message)
        self.ui.TXT_MAIN.setText(text)
        
    @staticmethod
    def show_text(parent, message, text):
        frm = FrmText(parent, message, text)
        frm.ui.BTNBOX_MAIN.button(QDialogButtonBox.Cancel).setVisible(False)
        
        return frm.exec_()
        
    @pyqtSlot()
    def on_BTNBOX_MAIN_accepted(self) -> None:
        """
        Signal handler:
        """
        self.accept()
    
    @pyqtSlot()
    def on_BTNBOX_MAIN_rejected(self) -> None:
        """
        Signal handler:
        """
        self.reject()



import traceback
import functools
from typing import Optional

from mhelper import string_helper, ImplementationError
from mhelper.generics_helper import DECORATOR_FUNCTION, DECORATOR, DECORATED


# NB. No global Qt imports.
#     mhelper must run without QT installed and should ideally defer loading of the QT library until it is actually used.


def exceptToGui() -> DECORATOR_FUNCTION:
    """
    DECORATOR
    
    Same as `exqtSlot` but without the `pyqtSlot` bit.
    """
    
    
    def true_decorator( fn ) -> DECORATOR:
        @functools.wraps( fn )
        def fn_wrapper( *args, **kwargs ) -> DECORATED:
            try:
                return fn( *args, **kwargs )
            except Exception as ex:
                show_exception( args[ 0 ], "Error", ex )


        # noinspection PyTypeChecker
        return fn_wrapper
    
    
    return true_decorator


def exqtSlot( *decorator_args ) -> DECORATOR_FUNCTION:
    """
    DECORATOR
    
    `pyqtSlot` is problematic in that if an exception occurs the application will silently exit.
    This decorator replaces `pyqtSlot` by adding not only the decorator, but a simple error handler. 
    """
    
    
    def true_decorator( fn ) -> DECORATOR:
        from PyQt5.QtCore import pyqtSlot
        return pyqtSlot( *decorator_args )( exceptToGui()( fn ) )
    
    
    return true_decorator


def browse_dir_on_textbox( textbox ) -> bool:
    """
    Opens a file browser, using a textbox to retrieve and store the filename.
    :type textbox: QLineEdit
    """
    owner = textbox.window()
    
    from PyQt5.QtWidgets import QFileDialog
    dir_name = str( QFileDialog.getExistingDirectory( owner, "Select directory", textbox.text(), QFileDialog.DontUseNativeDialog ) )
    
    if not dir_name:
        return False
    
    textbox.setText( dir_name )
    
    return True


def browse_dir( owner, existing = None ):
    from PyQt5.QtWidgets import QFileDialog
    return str( QFileDialog.getExistingDirectory( owner, "Select directory", existing, QFileDialog.DontUseNativeDialog ) )


def browse_open_on_textbox( textbox ) -> bool:
    """
    Opens a file browser, using a textbox to retrieve and store the filename.
    :type textbox: QLineEdit
    """
    from mhelper import file_helper
    
    owner = textbox.window()
    directory = file_helper.get_directory( textbox.text() )
    
    from PyQt5.QtWidgets import QFileDialog
    dir_name = str( QFileDialog.getOpenFileName( owner, "Select file", directory ) )
    
    if not dir_name:
        return False
    
    textbox.setText( dir_name )
    
    return True


def browse_save( parent, filter ):
    from mhelper import file_helper
    
    from PyQt5.QtWidgets import QFileDialog
    file_name, file_filter = QFileDialog.getSaveFileName( parent, "Save", "", filter, options = QFileDialog.DontUseNativeDialog )
    
    if file_name:
        if not file_helper.get_extension( file_name ):
            file_filter = get_extension_from_filter( file_filter )
            
            file_name += file_filter
    
    return file_name or None


def browse_open( parent, filter ):
    from PyQt5.QtWidgets import QFileDialog
    file_name, file_filter = QFileDialog.getOpenFileName( parent, "Open", "", filter, options = QFileDialog.DontUseNativeDialog )
    
    return file_name or None


def get_extension_from_filter( file_filter ):
    """
    Gets the first extension from a filter of the form
    
    Data ( *.txt ) --> .txt
    Data (*.txt) --> .txt
    Data (*.txt, *.csv) --> .txt
    Data (*.txt *.csv) --> .txt
    etc.
    """
    file_filter = file_filter.split( "(", 1 )[ 1 ].strip( " *" )
    
    for x in " ,)":
        if x in file_filter:
            file_filter = file_filter.split( x, 1 )[ 0 ]
    return file_filter


# noinspection PyUnusedLocal
def show_exception( owner, message = None, exception = None, traceback_ = None ) -> None:
    if not traceback_:
        traceback_ = traceback.format_exc()
    
    if isinstance( message, BaseException ):
        exception = message
        message = "Error"
    
    from PyQt5.QtWidgets import QMessageBox
    msg = QMessageBox()
    msg.setIcon( QMessageBox.Critical )
    msg.setText( str( message ) )
    
    if isinstance( exception, BaseException ):
        print( str( exception ) )
        print( traceback_ )
        msg.setInformativeText( str( exception ) )
        msg.setDetailedText( traceback_ )
    
    msg.exec_()


def to_check_state( value ):
    from PyQt5.QtCore import Qt
    if value is None:
        return Qt.PartiallyChecked
    elif value:
        return Qt.Checked
    else:
        return Qt.Unchecked


def from_check_state( value ):
    from PyQt5.QtCore import Qt
    if value == Qt.PartiallyChecked:
        return None
    elif value == Qt.Checked:
        return True
    elif value == Qt.Unchecked:
        return False
    else:
        from mhelper.exception_helper import SwitchError
        raise SwitchError( "from_check_state.value", value )


def move_treeview_items( source, destination, only_selected = True ):
    from PyQt5.QtWidgets import QTreeWidget
    assert isinstance( source, QTreeWidget )
    assert isinstance( destination, QTreeWidget )
    
    if only_selected:
        selected_items = source.selectedItems()
    else:
        selected_items = [ ]
        
        for index in range( source.topLevelItemCount() ):
            selected_items.append( source.topLevelItem( index ) )
    
    for item in selected_items:
        index = source.indexOfTopLevelItem( item )
        source.takeTopLevelItem( index )
        destination.addTopLevelItem( item )


class AnsiHtmlScheme:
    def __init__( self, values ):
        self.values = values
    
    
    def copy( self ):
        return AnsiHtmlScheme( dict( self.values ) )


def ansi_scheme_dark( fg = "#FFFFFF", bg = "#000000" ) -> AnsiHtmlScheme:
    """
    Creates a new ANSI scheme, defaulted to dark values.
    """
    #           code :  back,       fore,       style,      family
    return AnsiHtmlScheme( { 30 : ("", "#000000", "", ""),  # back::black
                             31 : ("", "#FF0000", "", ""),  # back::red
                             32 : ("", "#00FF00", "", ""),  # back::green
                             33 : ("", "#FFFF00", "", ""),  # back::yellow
                             34 : ("", "#0000FF", "", ""),  # back::blue
                             35 : ("", "#FF00FF", "", ""),  # back::magenta
                             36 : ("", "#00FFFF", "", ""),  # back::cyan
                             37 : ("", "#FFFFFF", "", ""),  # back::white
                             39 : ("", "*", "", ""),  # back::reset
                             90 : ("", "#808080", "", ""),  # back::light_black
                             91 : ("", "#FF8080", "", ""),  # back::light_red
                             92 : ("", "#80FF80", "", ""),  # back::light_green
                             93 : ("", "#FFFF80", "", ""),  # back::light_yellow
                             94 : ("", "#8080FF", "", ""),  # back::light_blue
                             95 : ("", "#FF80FF", "", ""),  # back::light_magenta
                             96 : ("", "#80FFFF", "", ""),  # back::light_cyan
                             97 : ("", "#FFFFFF", "", ""),  # back::light_white
                             40 : ("#000000", "", "", ""),  # back::black
                             41 : ("#FF0000", "", "", ""),  # back::red
                             42 : ("#00FF00", "", "", ""),  # back::green
                             43 : ("#FFFF00", "", "", ""),  # back::yellow
                             44 : ("#0000FF", "", "", ""),  # back::blue
                             45 : ("#FF00FF", "", "", ""),  # back::magenta
                             46 : ("#00FFFF", "", "", ""),  # back::cyan
                             47 : ("#FFFFFF", "", "", ""),  # back::white
                             49 : ("*", "", "", ""),  # fore::reset
                             100: ("#808080", "", "", ""),  # back::light_black
                             101: ("#FF8080", "", "", ""),  # back::light_red
                             102: ("#80FF80", "", "", ""),  # back::light_green
                             103: ("#FFFF80", "", "", ""),  # back::light_yellow
                             104: ("#8080FF", "", "", ""),  # back::light_blue
                             105: ("#FF80FF", "", "", ""),  # back::light_magenta
                             106: ("#80FFFF", "", "", ""),  # back::light_cyan
                             107: ("#FFFFFF", "", "", ""),  # back::light_white
                             0  : ("*", "*", "*", ""),  # style::reset_all
                             1  : ("", "", "bold", ""),  # style::bright
                             2  : ("", "", "italic", ""),  # style::dim
                             22 : ("", "", "normal", ""),  # style::normal
                             -1 : (bg, fg, "normal", "sans-serif"),  # special code: defaults
                             -2 : ("", "#00C0FF", "", "monospace"),  # quote start
                             -3 : ("", "*", "", "*"),  # end
                             -4 : ("", "", "", "monospace"),  # table start
                             -5 : ("", "", "", "*"),  # end
                             } )


def ansi_scheme_light( fg = "#000000", bg = "#FFFFFF" ) -> AnsiHtmlScheme:
    """
    Creates a new ANSI scheme, defaulted to light values.
    """
    #           code :  back,       fore,       style,      family
    return AnsiHtmlScheme( { 30 : ("", "#000000", "", ""),  # back::black
                             31 : ("", "#C00000", "", ""),  # back::red
                             32 : ("", "#00C000", "", ""),  # back::green
                             33 : ("", "#C0C000", "", ""),  # back::yellow
                             34 : ("", "#0000C0", "", ""),  # back::blue
                             35 : ("", "#C000C0", "", ""),  # back::magenta
                             36 : ("", "#00C0C0", "", ""),  # back::cyan
                             37 : ("", "#C0C0C0", "", ""),  # back::white
                             39 : ("", "*", "", ""),  # back::reset
                             90 : ("", "#C0C0C0", "", ""),  # back::light_black
                             91 : ("", "#FF0000", "", ""),  # back::light_red
                             92 : ("", "#00FF00", "", ""),  # back::light_green
                             93 : ("", "#FFFF00", "", ""),  # back::light_yellow
                             94 : ("", "#0000FF", "", ""),  # back::light_blue
                             95 : ("", "#FF00FF", "", ""),  # back::light_magenta
                             96 : ("", "#00FFFF", "", ""),  # back::light_cyan
                             97 : ("", "#FFFFFF", "", ""),  # back::light_white
                             40 : ("#000000", "", "", ""),  # back::black
                             41 : ("#C00000", "", "", ""),  # back::red
                             42 : ("#00C000", "", "", ""),  # back::green
                             43 : ("#C0C000", "", "", ""),  # back::yellow
                             44 : ("#0000C0", "", "", ""),  # back::blue
                             45 : ("#C000C0", "", "", ""),  # back::magenta
                             46 : ("#00C0C0", "", "", ""),  # back::cyan
                             47 : ("#C0C0C0", "", "", ""),  # back::white
                             49 : ("*", "", "", ""),  # fore::reset
                             100: ("#C0C0C0", "", "", ""),  # back::light_black
                             101: ("#FF0000", "", "", ""),  # back::light_red
                             102: ("#00FF00", "", "", ""),  # back::light_green
                             103: ("#FFFF00", "", "", ""),  # back::light_yellow
                             104: ("#0000FF", "", "", ""),  # back::light_blue
                             105: ("#FF00FF", "", "", ""),  # back::light_magenta
                             106: ("#00FFFF", "", "", ""),  # back::light_cyan
                             107: ("#FFFFFF", "", "", ""),  # back::light_white
                             0  : ("*", "*", "*", ""),  # style::reset_all
                             1  : ("", "", "bold", ""),  # style::bright
                             2  : ("", "", "italic", ""),  # style::dim
                             22 : ("", "", "normal", ""),  # style::normal
                             -1 : (bg, fg, "normal", "sans-serif"),  # special code: defaults
                             -2 : ("", "#0080C0", "", "Consolas, monospace"),  # quote start
                             -3 : ("", "*", "", "*"),  # end
                             -4 : ("", "", "", "Consolas, monospace"),  # table start
                             -5 : ("", "", "", "*"),  # end
                             } )


def ansi_scheme_get_style( lookup: Optional[ AnsiHtmlScheme ] = None ):
    lookup = lookup or ansi_scheme_dark()
    values = lookup.values[ -1 ]
    return 'background:{}; color:{}; font-style:{}; font-family:{}'.format( values[ 0 ], values[ 1 ], values[ 2 ], values[ 3 ] )


def ansi_to_html( text: Optional[ str ], lookup: Optional[ AnsiHtmlScheme ] = None ) -> str:
    if text is None:
        return ""
    
    from mhelper.string_parser import StringParser
    html = [ ]
    
    lookup = lookup.values if lookup is not None else ansi_scheme_dark().values
    default_style = lookup[ -1 ]
    current_style = list( default_style )
    
    text = string_helper.highlight_quotes( text, "`", "`", "\033[-2m", "\033[-3m" )
    text = string_helper.highlight_quotes( text, "┌", "┘", "\033[-4m", "\033[-5m" )
    text = string_helper.highlight_quotes( text, "«", "»", "\033[-2m", "\033[-3m" )
    
    text = text.replace( "\n* ", "\n• " )
    text = text.replace( "\n", '<br/>' )
    text = text.replace( " ", ' ' )
    
    parser = StringParser( text )
    
    html.append( '<span style="background:{}; color:{}; font-style:{}; font-family:{};">'.format( current_style[ 0 ], current_style[ 1 ], current_style[ 2 ], current_style[ 3 ] ) )
    iterations = 0
    
    ANSI_ESCAPE = '\033['
    
    while not parser.end():
        iterations += 1
        
        if iterations >= 10000:
            raise ImplementationError( "Possible infinite loop in mhelper.qt_gui_helper.ansi_to_html. Internal error. Possible solutions: Use a shorter string. Check the lookup syntax." )
        
        
        html.append( parser.read_past( ANSI_ESCAPE ) )
        
        if parser.end():
            break
        
        code = int( parser.read_past( "m" ) )
        new_style = lookup[ code ]
        
        for index in range( len( current_style ) ):
            if new_style[ index ]:
                if new_style[ index ] == "*":
                    current_style[ index ] = default_style[ index ]
                else:
                    current_style[ index ] = new_style[ index ]
        
        html.append( '<span style="background:{}; color:{}; font-style:{}; font-family:{};">'.format( current_style[ 0 ], current_style[ 1 ], current_style[ 2 ], current_style[ 3 ] ) )
    
    html.append( '</span>' )
    
    result = "".join( html )
    
    txx = [ "ANSI", text, "HTML", result ]
    from mhelper import file_helper
    file_helper.write_all_text( "temp-del-me.txt", "\n".join( txx ) )
    
    return result


__author__ = "Martin Rusilowicz"


class QtMutex:
    """
    A `QMutex` wrapped in Python.
    
    This is a `QMutex` and requires QT.
    - It is only for the GUI!
    - - Plugins should *not* use this!
    
    Usage:

        ```
        m = QtMutex()
        
        with m:
           . . .
           
        ```
           
    """
    
    
    # TODO: This shouldn't even live in this folder.
    
    def __init__( self ):
        from PyQt5.QtCore import QMutex
        self._mutex = QMutex()
    
    
    def __enter__( self ):
        self._mutex.lock()
        return self
    
    
    def __exit__( self, exc_type, exc_val, exc_tb ):
        self._mutex.unlock()


def suppress_debugging():
    # Disable PyQT logging to console since we use console for other stuff, plus, it's irritating
    # noinspection PyUnusedLocal
    def __message_handler( msg_type, msg_log_context, msg_string ):
        pass
    
    
    from PyQt5 import QtCore
    QtCore.qInstallMessageHandler( __message_handler )



"""
Contains the `ComponentFinder` class.
"""
from typing import List

class ComponentFinder:
    """
    Allows one to find connected components given a graph-like data-set without having to actually construct the graph. 
    
    Usage
    -----
    ```
    component_finder = ComponentFinder()
    
    component_finder.join(1, 2)
    component_finder.join(2, "Eggs")
    component_finder.join("Eggs", object)
    
    component_finder.tabulate()
    ```
    """
    
    
    def __init__( self ):
        """
        CONSTRUCTOR
        Nothing special.
        """
        self.__table = { }
        self.__next = 0
        self.__valid = set()
    
    
    def join( self, a, b ):
        """
        Sets the nodes `a` and `b` as being in the same component (i.e. that they share an edge).
        
        :param a: 
        :param b: 
        :return: 
        """
        # get existing components
        ac = self.__table.get( a )
        bc = self.__table.get( b )
        
        if ac is not None:
            if bc is not None:
                if ac != bc:
                    # we know now that ac and bc are actually the same!
                    self.__substitute( bc, ac )
            else:
                # b has no component, so it belongs to a
                self.__table[ b ] = ac
        else:
            if bc is not None:
                self.__table[ a ] = bc
            else:
                self.__table[ a ] = self.__next
                self.__table[ b ] = self.__next
                self.__valid.add( self.__next )
                self.__next += 1
    
    
    def tabulate( self ) -> List[ List[ object ] ]:
        """
        Obtains the result as a list of lists.
        Each list in the result represents a component and its contents the contents of that component.
        
        :return:    List[ List[ object ] ]
        """
        order = list( self.__valid )
        new_order = { }
        
        for index, value in enumerate( order ):
            new_order[ value ] = index
        
        for k, v in list( self.__table.items() ):
            vv = new_order[ v ]
            self.__table[ k ] = vv
        
        self.__valid = set( new_order.values() )
        
        self.__next = len( self.__valid )
        
        return self.__create_groups()
    
    
    def __create_groups( self ):
        results = [ ]
        for i in range( self.__next ):
            results.append( [ ] )
        for k, v in self.__table.items():
            results[ v ].append( k )
        return results
    
    
    def __substitute( self, original_number, replacement_number ):
        fix = [ ]
        
        for k, v in self.__table.items():
            if v == original_number:
                fix.append( k )
        
        for k in fix:
            self.__table[ k ] = replacement_number
        
        self.__valid.remove( original_number )



"""
Load and save functions
"""
import warnings
from typing import Optional, TypeVar, Union, cast


T = TypeVar("T")


CAUSE_ERROR = object # Sentinel object for load_* functions

def load_npy( file_name ):
    import numpy
    result = numpy.load( file_name, allow_pickle = False, fix_imports = False )
    
    if file_name.upper().endswith( ".NPZ" ):
        result = result[ "data" ]
    
    return result


def load_npz( file_name ):
    import numpy
    result = numpy.load( file_name, allow_pickle = False, fix_imports = False )
    result = result[ "data" ]
    
    return result

def save_bitarray(file_name : str, value) -> None:
    """
    Saves a bit array
    
    :type value: bitarray
    
    :param file_name: File name
    :param value: Value to save
    """
    from bitarray import bitarray
    from mhelper.exception_helper import SubprocessError
    from mhelper import exception_helper, file_helper
    
    exception_helper.assert_instance( "save_bitarray::value", value, bitarray )
    assert isinstance(value, bitarray) 
    
    try:
        with open(file_name, "wb") as file_out:
            value.tofile(file_out)
    except TypeError:
        # Stupid "open file expected" error on OSX (due to bug writing large files - fallback to manual implementation)
        _save_bytes_manually( file_name, value.tobytes() )
    except Exception as ex:
        raise SubprocessError("Failed to write bitarray of length {} to «{}».".format(len(value), file_name)) from ex
    
    size = float( file_helper.file_size( file_name ) )
    expected = len(value)
    
    if size < expected / 8.0:
        raise ValueError("Saved file is shorter ({} bytes or {} bits) than the originating bitarray ({} bytes or {} bits).".format(size, size * 8, expected / 8, expected))
    
def _save_bytes_manually( file_name: str, value: bytes ):
    """
    Fallback function used by `save_bitarray`.
    """
    warnings.warn( "Save bitarray failed. This is probably due to an error in your `bitarray` library. The file will be saved incrementally but this will take longer.", UserWarning )
    BATCH_SIZE = 100000
    cursor = 0
    length = len( value )
    
    with open( file_name, "wb" ) as file_out: #note that writing large arrays on OSX is probably the problem, we can't just dump the bytes
        while cursor < length:
            next = min( cursor + BATCH_SIZE, length )
            slice = value[ cursor:next ]
            file_out.write( slice ) 
            cursor = next
            
def _read_bytes_manually(file_name:str)->bytes:
    BATCH_SIZE = 100000
    b = bytearray()
    
    with open( file_name, "rb" ) as file_in:
        while True:
            buf = file_in.read( BATCH_SIZE )
            
            if len( buf ) == 0:
                break
            
            b.extend( buf )
    
    return bytes( b )

def load_bitarray(file_name : str):
    """
    Loads a bitarray
    :param file_name: File to read 
    :return: bitarray. Note this may be padded with missing bits. 
    """
    from bitarray import bitarray
    
    result = bitarray()
    
    try:
        with open(file_name, 'rb') as file_in:
            result.fromfile(file_in)
    except SystemError: #OSX error
        result = bitarray()
        result.frombytes(_read_bytes_manually(file_name))
        assert len(result)!=0
        
    return result


def save_npy( file_name, value ):
    import numpy
    
    numpy.save( file_name, value, allow_pickle = False, fix_imports = False )


def save_npz( file_name, value ):
    import numpy
    
    numpy.savez_compressed( file_name, data = value )


def load_binary( file_name : str, default : Optional[object] = CAUSE_ERROR ):
    """
    Loads a binary file
    
    :param file_name: Filename to load  
    :param default:     Whether to default the values (if not specified or `cause_error` an exception is raised) 
    :return: Loaded file, or `default` (if specified) 
    """
    try:
        import pickle
        
        with open( file_name, "rb" ) as file:
            return pickle.load( file, fix_imports = False )
    except Exception:
        if default is CAUSE_ERROR:
            raise
        
        return default
    except BaseException as ex:
        warnings.warn("Loading the binary file «{}» caused a «{}» exception.".format(file_name, type(ex).__name__), UserWarning)
        raise


def save_binary( file_name, value ):
    import pickle
    
    with open( file_name, "wb" ) as file:
        pickle.dump( value, file, fix_imports = False )


def load_json( file_name, keys = True ):
    from mhelper.file_helper import read_all_text
    import jsonpickle
    
    text = read_all_text( file_name )
    
    return jsonpickle.decode( text, keys = keys )


def save_json( file_name, value, keys = True ):
    from mhelper.file_helper import write_all_text
    import jsonpickle
    
    write_all_text( file_name, jsonpickle.encode( value, keys = keys ) )


def default_values( target: T, default: Optional[ Union[ T, type ] ] = None ) -> T:
    if default is None:
        if target is None:
            raise ValueError("Cannot set the defaults for the value because both the value and the defaults are `None`, so neither can be inferred.")
        
        default = type(target)
        
    if isinstance(default, type):
        default = default()
    
    if target is None:
        return default
    
    if isinstance(target, list):
        return cast(T, target)
    
    # TODO: PUT THIS BACK IN! Removed because for some reason we get a different class when deserialising from the IDE, but this is a problem!
    # if type(target) is not type(default):
    #     raise ValueError("Attempting to set the defaults for the value «{}», of type «{}», but the value provided, «{}», of type «{}», is not compatible with this.".format(target, type(target), default, type(default)))
    
    for k, v in default.__dict__.items():
        if k.startswith("_"):
            continue
            
        if k not in target.__dict__:
            target.__dict__[k] = v

    to_delete = []
            
    for k in target.__dict__.keys():
        if k not in default.__dict__:
            to_delete.append(k)

    for k in to_delete:
        del target.__dict__[k]
            
    return target


from random import randint

from PyQt5.QtGui import QColor, QPen, QBrush


class Colours:
    BLACK            = QColor( 0, 0, 0 )
    DARK_GRAY        = QColor( 64, 64, 64 )
    GRAY             = QColor( 128, 128, 128 )
    LIGHT_GRAY       = QColor( 192, 192, 192 )
    WHITE            = QColor( 255, 255, 255 )
    
    RED              = QColor( 255, 0, 0 )
    GREEN            = QColor( 0, 255, 0 )
    BLUE             = QColor( 0, 0, 255 )
    
    CYAN             = QColor( 0, 255, 255 )
    MAGENTA          = QColor( 255, 0, 255 )
    YELLOW           = QColor( 255, 255, 0 )
    
    ORANGE           = QColor( 255, 128, 0 )
    FUSCHIA          = QColor( 255, 0, 128 )
    LIME             = QColor( 128, 255, 0 )
    SPRING_GREEN     = QColor( 0, 255, 128 )
    PURPLE           = QColor( 128, 0, 255 )
    DODGER_BLUE      = QColor( 128, 0, 255 )
    
    DARK_RED         = QColor( 128, 0, 0 )
    DARK_GREEN       = QColor( 0, 128, 0 )
    DARK_BLUE        = QColor( 0, 0, 128 )
    
    DARK_CYAN        = QColor( 0, 128, 128 )
    DARK_MAGENTA     = QColor( 128, 0, 128 )
    DARK_YELLOW      = QColor( 128, 128, 0 )
    
    DARK_ORANGE      = QColor( 128, 64, 0 )
    DARK_FUSCHIA     = QColor( 128, 0, 64 )
    DARK_LIME        = QColor( 64, 128, 0 )
    DARK_SPRINGGREEN = QColor( 0, 128, 64 )
    DARK_PURPLE      = QColor( 64, 0, 128 )
    DARK_DODGERBLUE  = QColor( 64, 0, 128 )
    
    LIGHT_RED        = QColor( 255, 128, 128 )
    LIGHT_GREEN      = QColor( 128, 255, 128 )
    LIGHT_BLUE       = QColor( 128, 128, 255 )
    
    LIGHT_CYAN       = QColor( 128, 255, 255 )
    LIGHT_MAGENTA    = QColor( 255, 0, 255 )
    LIGHT_YELLOW     = QColor( 255, 255, 128 )


class Pens:
    BLACK            = QPen( Colours.BLACK )
    DARK_GRAY        = QPen( Colours.DARK_GRAY )
    GRAY             = QPen( Colours.GRAY )
    LIGHT_GRAY       = QPen( Colours.LIGHT_GRAY )
    WHITE            = QPen( Colours.WHITE )
    RED              = QPen( Colours.RED )
    GREEN            = QPen( Colours.GREEN )
    BLUE             = QPen( Colours.BLUE )
    CYAN             = QPen( Colours.CYAN )
    MAGENTA          = QPen( Colours.MAGENTA )
    YELLOW           = QPen( Colours.YELLOW )
    ORANGE           = QPen( Colours.ORANGE )
    FUSCHIA          = QPen( Colours.FUSCHIA )
    LIME             = QPen( Colours.LIME )
    SPRING_GREEN     = QPen( Colours.SPRING_GREEN )
    PURPLE           = QPen( Colours.PURPLE )
    DODGER_BLUE      = QPen( Colours.DODGER_BLUE )
    DARK_RED         = QPen( Colours.DARK_RED )
    DARK_GREEN       = QPen( Colours.DARK_GREEN )
    DARK_BLUE        = QPen( Colours.DARK_BLUE )
    DARK_CYAN        = QPen( Colours.DARK_CYAN )
    DARK_MAGENTA     = QPen( Colours.DARK_MAGENTA )
    DARK_YELLOW      = QPen( Colours.DARK_YELLOW )
    DARK_ORANGE      = QPen( Colours.DARK_ORANGE )
    DARK_FUSCHIA     = QPen( Colours.DARK_FUSCHIA )
    DARK_LIME        = QPen( Colours.DARK_LIME )
    DARK_SPRINGGREEN = QPen( Colours.DARK_SPRINGGREEN )
    DARK_PURPLE      = QPen( Colours.DARK_PURPLE )
    DARK_DODGERBLUE  = QPen( Colours.DARK_DODGERBLUE )
    LIGHT_RED        = QPen( Colours.LIGHT_RED )
    LIGHT_GREEN      = QPen( Colours.LIGHT_GREEN )
    LIGHT_BLUE       = QPen( Colours.LIGHT_BLUE )
    LIGHT_CYAN       = QPen( Colours.LIGHT_CYAN )
    LIGHT_MAGENTA    = QPen( Colours.LIGHT_MAGENTA )
    LIGHT_YELLOW     = QPen( Colours.LIGHT_YELLOW )

class Brushes:
    BLACK            = QBrush( Colours.BLACK )
    DARK_GRAY        = QBrush( Colours.DARK_GRAY )
    GRAY             = QBrush( Colours.GRAY )
    LIGHT_GRAY       = QBrush( Colours.LIGHT_GRAY )
    WHITE            = QBrush( Colours.WHITE )
    RED              = QBrush( Colours.RED )
    GREEN            = QBrush( Colours.GREEN )
    BLUE             = QBrush( Colours.BLUE )
    CYAN             = QBrush( Colours.CYAN )
    MAGENTA          = QBrush( Colours.MAGENTA )
    YELLOW           = QBrush( Colours.YELLOW )
    ORANGE           = QBrush( Colours.ORANGE )
    FUSCHIA          = QBrush( Colours.FUSCHIA )
    LIME             = QBrush( Colours.LIME )
    SPRING_GREEN     = QBrush( Colours.SPRING_GREEN )
    PURPLE           = QBrush( Colours.PURPLE )
    DODGER_BLUE      = QBrush( Colours.DODGER_BLUE )
    DARK_RED         = QBrush( Colours.DARK_RED )
    DARK_GREEN       = QBrush( Colours.DARK_GREEN )
    DARK_BLUE        = QBrush( Colours.DARK_BLUE )
    DARK_CYAN        = QBrush( Colours.DARK_CYAN )
    DARK_MAGENTA     = QBrush( Colours.DARK_MAGENTA )
    DARK_YELLOW      = QBrush( Colours.DARK_YELLOW )
    DARK_ORANGE      = QBrush( Colours.DARK_ORANGE )
    DARK_FUSCHIA     = QBrush( Colours.DARK_FUSCHIA )
    DARK_LIME        = QBrush( Colours.DARK_LIME )
    DARK_SPRINGGREEN = QBrush( Colours.DARK_SPRINGGREEN )
    DARK_PURPLE      = QBrush( Colours.DARK_PURPLE )
    DARK_DODGERBLUE  = QBrush( Colours.DARK_DODGERBLUE )
    LIGHT_RED        = QBrush( Colours.LIGHT_RED )
    LIGHT_GREEN      = QBrush( Colours.LIGHT_GREEN )
    LIGHT_BLUE       = QBrush( Colours.LIGHT_BLUE )
    LIGHT_CYAN       = QBrush( Colours.LIGHT_CYAN )
    LIGHT_MAGENTA    = QBrush( Colours.LIGHT_MAGENTA )
    LIGHT_YELLOW     = QBrush( Colours.LIGHT_YELLOW )


def interpolate_float( a: float, b: float, fb: float ) -> float:
    return a + (b - a) * fb


def interpolate_int( a: int, b: int, fb: float ) -> int:
    return int( interpolate_float( a, b, fb ) )


def interpolate_colours( a: QColor, b: QColor, fb: float ) -> QColor:
    ca = interpolate_int( a.alpha(), b.alpha(), fb )
    cr = interpolate_int( a.red(), b.red(), fb )
    cg = interpolate_int( a.green(), b.green(), fb )
    cb = interpolate_int( a.blue(), b.blue(), fb )
    return QColor( cr, cg, cb, ca )

def average_colour( the_list ):
    r = sum( x.red() for x in the_list ) // len( the_list )
    g = sum( x.green() for x in the_list ) // len( the_list )
    b = sum( x.blue() for x in the_list ) // len( the_list )
    return QColor( r, g, b )


def random_colour():
    return QColor(randint(0,255),randint(0,255),randint(0,255))


import inspect
from typing import List, Optional, Union, Callable, TypeVar, Type, Dict, cast, GenericMeta
from mhelper.generics_helper import MAnnotation, MAnnotationFactory
from mhelper.special_types import MOptional, MUnion


T = TypeVar( "T" )

_TUnion = type( Union[int, str] )


class AnnotationInspector:
    """
    Class to inspect PEP484 generics.
    """
    
    
    def __init__( self, annotation: object ):
        """
        CONSTRUCTOR
        :param annotation: `type` to inspect 
        """
        self.value = annotation
    
    
    def __str__( self ) -> str:
        """
        Returns the underlying type string
        """
        return str( self.value )
    
    
    @property
    def is_mannotation( self ):
        """
        Is this an instance of `MAnnotation`?
        """
        return isinstance( self.value, MAnnotation )
    
    
    def is_mannotation_of( self, parent: MAnnotationFactory ):
        """
        Is this an instance of `MAnnotation`, specifically a `specific_type` derivative?
        """
        if not self.is_mannotation:
            return False
        
        assert isinstance( self.value, MAnnotation )
        return self.value.factory is parent
    
    
    @property
    def mannotation( self ) -> MAnnotation:
        """
        Returns the MAnnotation object, if this is an MAnnotation, otherwise raises an error.
        
        :except TypeError: Not an MAnnotation.
        """
        if not self.is_mannotation:
            raise TypeError( "«{}» is not an MAnnotation[T].".format( self ) )
        
        return cast( MAnnotation, self.value )
    
    
    @property
    def mannotation_arg( self ):
        """
        If this is an instance of `MAnnotation`, return the underlying type, otherwise, raise an error.
        """
        if not self.is_mannotation:
            raise TypeError( "«{}» is not an MAnnotation[T].".format( self ) )
        
        assert isinstance( self.value, MAnnotation )
        return self.value.child
    
    
    @property
    def is_glist( self ) -> bool:
        """
        Is this a `List[T]`?
        
        (note: this does not include `list` or `List` with no `T`)
        """
        return isinstance(self.value, type) and issubclass(cast(type, self.value), List) and self.value is not list and hasattr( self.value, "__args__" )
    
    
    @property
    def glist_type( self ):
        """
        Gets the T in List[T]. Otherwise raises `TypeError`.
        """
        if not self.is_glist:
            raise TypeError( "«{}» is not a List[T].".format( self ) )
        
        # noinspection PyUnresolvedReferences
        return self.value.__args__[0]
    
    
    @property
    def is_union( self ) -> bool:
        """
        Is this a `Union[T, ...]`?
        Is this a `MUnion[T, ...]`?
        """
        return isinstance( self.value, _TUnion ) or self.is_mannotation_of( MUnion )
    
    
    def is_directly_below( self, upper_class: type ) -> bool:
        """
        Is `self.value` a sub-class of `lower_class`?
        """
        if not self.is_type:
            return False
            
        try:
            return issubclass( cast( type, self.value ), upper_class )
        except TypeError as ex:
            raise TypeError("Cannot determine if «{}» is directly below «{}» because `issubclass({}, {})` returned an error.".format(self, upper_class, self, upper_class)) from ex
    
    def is_directly_above( self, lower_class: type ) -> bool:
        """
        Is `lower_class` a sub-class of `self.value`?
        """
        if not self.is_type:
            return False
        
        if self.is_glist:
            # Special case for List[T]
            return issubclass(lower_class, list)
            
        try:
            return issubclass( lower_class, cast( type, self.value ) )
        except TypeError as ex:
            raise TypeError("Cannot determine if «{}» is directly above «{}» because `issubclass({}, {})` returned an error.".format(self, lower_class, lower_class, self)) from ex
    
    
    def get_directly_below( self, upper_class: type ) -> Optional[type]:
        """
        This is the same as `is_directly_below`, but returns the true `type` (`self.value`) if `True`.
        """
        if self.is_directly_below( upper_class ):
            return cast( type, self.value )
    
    
    def get_directly_above( self, lower_class: type ) -> Optional[type]:
        """
        This is the same as `is_directly_above`, but returns the true `type` (`self.value`) if `True`.
        """
        if self.is_directly_above( lower_class ):
            return cast( type, self.value )
    
    
    def is_indirectly_below( self, upper_class: type ) -> bool:
        """
        Is `self.value` a sub-class of `upper_class`, or an annotation enclosing a class that is a sub-class of `upper_class`? 
        """
        return self.get_indirectly_below( upper_class ) is not None
    
    
    def is_indirectly_above( self, lower_class: type ) -> bool:
        """
        Is `lower_class` a sub-class of `self.value`, or a sub-class of an annotation enclosed within `self.value`?
        """
        return self.get_indirectly_above( lower_class ) is not None
    
    
    def get_indirectly_above( self, lower_class: type ) -> Optional[type]:
        """
        This is the same as `is_indirectly_below`, but returns the true `type` that is above `lower_class`.
        """
        return self.__get_indirectly( lower_class, AnnotationInspector.is_directly_above )
    
    
    def get_indirectly_below( self, upper_class: type ) -> Optional[type]:
        """
        This is the same as `is_indirectly_above`, but returns the true `type` that iis below `upper_class`.
        """
        return self.__get_indirectly( upper_class, AnnotationInspector.is_directly_below )
    
    
    def __get_indirectly( self, query: type, predicate: "Callable[[AnnotationInspector, type],bool]" ) -> Optional[object]:
        """
        Checks inside all `Unions` and `MAnnotations` until the predicate matches, returning the type (`self.value`) when it does.
        """
        if predicate( self, query ):
            return self.value
        
        if self.is_union:
            for arg in self.union_args:
                arg_type = AnnotationInspector( arg ).__get_indirectly( query, predicate )
                
                if arg_type is not None:
                    return arg_type
        
        if self.is_mannotation:
            annotation_type = AnnotationInspector( self.mannotation_arg ).__get_indirectly( query, predicate )
            
            if annotation_type is not None:
                return annotation_type
        
        return None
    
    
    @property
    def union_args( self ) -> List[type]:
        """
        Returns the list of Union parameters `[...]`.
        """
        if not self.is_union:
            raise TypeError( "«{}» is not a Union[T].".format( self ) )
        
        # noinspection PyUnresolvedReferences
        if self.is_mannotation_of( MUnion ):
            return self.mannotation_arg
        else:
            return cast( _TUnion, self.value ).__args__
    
    
    @property
    def is_optional( self ) -> bool:
        """
        If a `Union[T, U]` where `None` in `T`, `U`.
        """
        if self.is_mannotation_of( MOptional ):
            return True
        
        if not self.is_union:
            return False
        
        if len( self.union_args ) == 2 and None in self.union_args:
            return True
        
        return False
    
    
    @property
    def is_multi_optional( self ) -> bool:
        """
        If a `Union[...]` with `None` in `...`
        """
        if self.is_mannotation_of( MOptional ):
            return True
        
        if not self.is_union:
            return False
        
        if None in self.union_args:
            return True
        
        return False
    
    
    @property
    def optional_types( self ) -> Optional[List[type]]:
        """
        Returns `...` in a `Union[None, ...]`, otherwise raises `TypeError`.
        """
        if self.is_mannotation_of( MOptional ):
            return [self.mannotation_arg]
        
        if not self.is_union:
            raise TypeError( "«{}» is not a Union[T].".format( self ) )
        
        union_params = self.union_args
        
        if type( None ) not in union_params:
            raise TypeError( "«{}» is not a Union[...] with `None` in `...`.".format( self ) )
        
        union_params = list( self.union_args )
        union_params.remove( type( None ) )
        return union_params
    
    
    @property
    def optional_type( self ) -> Optional[type]:
        """
        Returns `T` in a `Union[T, U]` (i.e. an `Optional[T]`). Otherwise raises `TypeError`.
        """
        t = self.optional_types
        
        if len( t ) == 1:
            return t[0]
        else:
            raise TypeError( "«{}» is not a Union[T, None] (i.e. an Optional[T]).".format( self ) )
    
    
    @property
    def type_or_optional_type( self ) -> Optional[type]:
        """
        If this is an `Optional[T]`, returns `T`.
        If this is a `T`, returns `T`.
        Otherwise returns `None`.
        """
        if self.is_optional:
            return self.optional_type
        elif self.is_type:
            assert isinstance( self.value, type )
            return self.value
        else:
            return None
    
    
    @property
    def safe_type( self ) -> Optional[type]:
        """
        If this is a `T`, returns `T`, else returns `None`.
        """
        if self.is_type:
            assert isinstance( self.value, type )
            return self.value
        else:
            return None
    
    
    @property
    def is_type( self ):
        """
        Returns if my `type` is an actual `type`, not an annotation object like `Union`.
        """
        return isinstance( self.value, type ) 
    
    
    @property
    def name( self ):
        """
        Returns the type name
        """
        if not self.is_type:
            raise TypeError( "«{}» is not a <type>.".format( self ) )
        
        return self.value.__name__
    
    
    def is_type_or_optional_type_subclass( self, u: type ):
        """
        Returns if `type_or_optional_type` is a subclass `u`.
        """
        return issubclass( self.type_or_optional_type, u )
    
    
    def is_viable_instance( self, value ):
        """
        Returns `is_viable_subclass` on the value's type.
        """
        return self.is_indirectly_above( type( value ) )


def as_delegate( x: Union[T, Callable[[], T]], t: Type[T] ) -> Callable[[], T]:
    """
    If `x` is a `t`, returns a lambda returning `x`, otherwise, assumes `x` is already a lambda and returns `x`.
    This is the opposite of `dedelegate`.
    """
    if isinstance( x, t ):
        return (lambda x: lambda: x)( x )
    else:
        return x


def defunction( x ):
    """
    If `x` is a function or a method, calls `x` and returns the result.
    Otherwise, returns `x`.
    """
    if inspect.isfunction( x ) or inspect.ismethod( x ):
        return x()
    else:
        return x


def dedelegate( x: Union[T, Callable[[], T]], t: Type[T] ) -> T:
    """
    If `x` is not a `t`, calls `x` and returns the result.
    Otherwise, returns `x`.
    This is the opposite of `as_delegate`.
    """
    if not isinstance( x, t ):
        x = x()
    
    return x


def public_dict( d: Dict[str, object] ) -> Dict[str, object]:
    """
    Yields the public key-value pairs.
    """
    r = { }
    
    for k, v in d.items():
        if not k.startswith( "_" ):
            r[k] = v
    
    return r





def convert_file(in_filename, out_filename, in_format, out_format):
    from Bio import AlignIO
    
    with open(in_filename, "rU") as input_handle:
        with open(out_filename, "w") as output_handle:
            alignments = AlignIO.parse(input_handle,in_format)
            AlignIO.write(alignments, output_handle, out_format)



from typing import Tuple


class BatchList:
    def __init__(self, data, batch_size):
        self.batch = batch_size
        self.data = list(data)
        
    def take(self):
        result = self.data[0:self.batch]
        del self.data[0:self.batch]
        return result
    
    def __bool__(self):
        return bool(self.data)
    
    def __len__(self):
        return len(self.data)
    
    def __iter__(self):
        while self:
            yield self.take()


def divide_workload(index : int, count : int, quantity : int) -> Tuple[int, int]:
    partition_size = quantity / count
    
    start = index * partition_size
    
    if index == count - 1:
        end = quantity
    else:
        end  = int(start + partition_size)
        
    return int(start), end
    


from typing import Optional, Callable


class ManagedWith:
    """
    Wraps a (presumably pooled) object.
    
    Usage:

        ```
        with ManagedWith(...):
            ...
        ```
        
    Example (database provider):
    
        def get_database():
            return ManagedWith(pool.pop(), pool.push)
        
    Example (from `Plugin`):
        
        ```
        with get_database() as db:
            db.execute("RETURN 1")
        ```
    """
    
    
    def YOURE_USING_THE_WRONG_THING( self ):
        """
        Message to an IDE user telling them they're using the wrong thing! (should be using `with`)
        """
        pass
    
    
    def __init__( self,
                  target: Optional[ object ] = None,
                  callback: Callable[ [ Optional[ object ] ], None ] = None,
                  call_into: Callable[ [ Optional[ object ] ], None ] = None,
                  obtain: Callable[ [ ], object ] = None ):
        """
        :param target: Object to manage 
        :param callback: What to do when the caller has finished with the target 
        """
        self.__target = target
        self.__callback = callback
        self.__call_into = call_into
        self.__obtain = obtain
        
        if self.__obtain is not None and self.__target is not None:
            raise ValueError( "Cannot specify both 'obtain' and 'target' parameters." )
    
    
    def __enter__( self ):
        if self.__obtain is not None:
            self.__target = self.__obtain()
        
        if self.__call_into is not None:
            self.__call_into( self.__target )
        
        return self.__target
    
    
    def __exit__( self, exc_type, exc_val, exc_tb ):
        if self.__callback is not None:
            self.__callback( self.__target )
        
        if self.__obtain is not None:
            self.__target = None



from mhelper.array_helper import Indexer
from mhelper.batch_lists import BatchList
from mhelper.comment_helper import override, abstract, virtual, sealed, overrides, protected, ignore
from mhelper.disposal_helper import ManagedWith
from mhelper.exception_helper import NotSupportedError, SwitchError, LogicError, ImplementationError, NotFoundError
from mhelper.log_helper import Logger
from mhelper.proxy_helper import SimpleProxy, PropertySetInfo
from mhelper.qt_gui_helper import exceptToGui
from mhelper.reflection_helper import AnnotationInspector
from mhelper.generics_helper import MGeneric, MGenericMeta, MAnnotation, GenericString, NonGenericString, GenericStringMeta, MAnnotationFactory, MAnnotation, ByRef, MAnnotationArgs
from mhelper.special_types import Password, Filename, Dirname, EFileMode, MEnum, MFlags, HReadonly, MOptional, MUnion, FileNameAnnotation
from mhelper.string_helper import FindError

from mhelper import array_helper
from mhelper import batch_lists
from mhelper import bio_helper
from mhelper import comment_helper
from mhelper import component_helper
from mhelper import disposal_helper
from mhelper import event_helper
from mhelper import exception_helper
from mhelper import file_helper
from mhelper import generics_helper
from mhelper import io_helper
from mhelper import log_helper
from mhelper import maths_helper
from mhelper import print_helper
from mhelper import proxy_helper
from mhelper import qt_gui_helper
from mhelper import reflection_helper
from mhelper import string_helper
from mhelper import string_parser

__author__ = "Martin Rusilowicz"
__version__ = "1.0.1.23"



from typing import Callable, Optional


class PropertySetInfo:
    def __init__( self, source, name, value ):
        self.source = source
        self.name = name
        self.value = value


class SimpleProxy:
    def __init__( self, source_get: Callable[[], object] = None, source: object = None, watch: Optional[Callable[[PropertySetInfo], None]] = None ):
        if source_get is None:
            if source is None:
                raise ValueError( "Must specify either the «source_get» or the «source» parameter." )
            
            source_get = (lambda x: lambda: x)( source )
            
        object.__setattr__( self, "__source", source_get )
        object.__setattr__( self, "__watch", watch )
    
    
    def __getattribute__( self, name: str ) -> object:
        if name == "__class__":
            return SimpleProxy
        
        source = object.__getattribute__( self, "__source" )
        return getattr( source(), name )
    
    
    def __delattr__( self, name ):
        delattr( object.__getattribute__( self, "__source" )(), name )
    
    
    def __setattr__( self, name, value ):
        watch = object.__getattribute__( self, "__watch" )
        source = object.__getattribute__( self, "__source" )()
        setattr( source, name, value )
        
        if watch is not None:
            args = PropertySetInfo( source, name, value )
            watch( args )
    
    
    def __str__( self ) -> str:
        return str( object.__getattribute__( self, "__source" )() )
    
    
    def __repr__( self ) -> str:
        return repr( object.__getattribute__( self, "__source" )() )




# Special
RESET = "\033[0m"

# Styles
BOLD = "\033[1m"
DIM = "\033[2m"
ITALIC = "\033[3m"
UNDERLINE = "\033[4m"
BLINK = "\033[5m"
INVERSE = "\033[7m"
NORMAL = "\033[22m"

# Foreground
FORE_BLACK = "\033[30m"
FORE_RED = "\033[31m"
FORE_GREEN = "\033[32m"
FORE_YELLOW = "\033[33m"
FORE_BLUE = "\033[34m"
FORE_MAGENTA = "\033[35m"
FORE_CYAN = "\033[36m"
FORE_WHITE = "\033[37m"
FORE_RESET = "\033[39m"
FORE_BRIGHT_BLACK = "\033[1;30m"
FORE_BRIGHT_RED = "\033[1;31m"
FORE_BRIGHT_GREEN = "\033[1;32m"
FORE_BRIGHT_YELLOW = "\033[1;33m"
FORE_BRIGHT_BLUE = "\033[1;34m"
FORE_BRIGHT_MAGENTA = "\033[1;35m"
FORE_BRIGHT_CYAN = "\033[1;36m"
FORE_BRIGHT_WHITE = "\033[1;37m"

# Background
BACK_BLACK = "\033[40m"
BACK_RED = "\033[41m"
BACK_GREEN = "\033[42m"
BACK_YELLOW = "\033[43m"
BACK_BLUE = "\033[44m"
BACK_MAGENTA = "\033[45m"
BACK_CYAN = "\033[46m"
BACK_WHITE = "\033[47m"
BACK_RESET = "\033[49m"
BACK_LIGHT_BLACK = "\033[100m"
BACK_LIGHT_RED = "\033[101m"
BACK_LIGHT_GREEN = "\033[102m"
BACK_LIGHT_YELLOW = "\033[103m"
BACK_LIGHT_BLUE = "\033[104m"
BACK_LIGHT_MAGENTA = "\033[105m"
BACK_LIGHT_CYAN = "\033[106m"
BACK_LIGHT_WHITE = "\033[107m"

# Do not use
BACK_BRIGHT_BLACK = "\033[100m"
BACK_BRIGHT_RED = "\033[101m"
BACK_BRIGHT_GREEN = "\033[102m"
BACK_BRIGHT_YELLOW = "\033[103m"
BACK_BRIGHT_BLUE = "\033[104m"
BACK_BRIGHT_MAGENTA = "\033[105m"
BACK_BRIGHT_CYAN = "\033[106m"
BACK_BRIGHT_WHITE = "\033[107m"

# 24-bit colours
def fore( r, g, b ):
    return "\033[38;2;" + str( r ) + ";" + str( g ) + ";" + str( b ) + "m"

def back( r, g, b ):
    return "\033[48;2;" + str( r ) + ";" + str( g ) + ";" + str( b ) + "m"

# Text
FR = FORE_RED
FG = FORE_GREEN
FB = FORE_BLUE
FC = FORE_CYAN
FY = FORE_YELLOW
FM = FORE_MAGENTA
FW = FORE_WHITE
FK = FORE_BLACK

FBR = FORE_BRIGHT_RED
FBG = FORE_BRIGHT_GREEN
FBB = FORE_BRIGHT_BLUE
FBC = FORE_BRIGHT_CYAN
FBY = FORE_BRIGHT_YELLOW
FBM = FORE_BRIGHT_MAGENTA
FBW = FORE_BRIGHT_WHITE
FBK = FORE_BRIGHT_BLACK

#Background
BR = BACK_RED
BG = BACK_GREEN
BB = BACK_BLUE
BC = BACK_CYAN
BY = BACK_YELLOW
BM = BACK_MAGENTA
BW = BACK_WHITE
BK = BACK_BLACK

BBR = BACK_LIGHT_RED
BBG = BACK_LIGHT_GREEN
BBB = BACK_LIGHT_BLUE
BBC = BACK_LIGHT_CYAN
BBY = BACK_LIGHT_YELLOW
BBM = BACK_LIGHT_MAGENTA
BBW = BACK_LIGHT_WHITE
BBK = BACK_LIGHT_BLACK

# Styles
SB = BOLD
SD = DIM
SI = ITALIC
SU = UNDERLINE
SE = INVERSE



import unittest
from typing import Optional, Union

from mhelper import BatchList, array_helper, batch_lists, string_helper, AnnotationInspector




class string_helper_tests( unittest.TestCase ):
    def test_highlight_quotes( self ):
        self.assertEqual( string_helper.highlight_quotes( "hello 'world' hello's", "'", "'", "<", ">" ), "hello <world> hello's" )
    
    
    def test_fix_indents( self ):
        text_a = """alpha\n\n        beta\n            gamma\n                delta\n        epsilon\n    """
        text_b = """alpha\n\nbeta\n    gamma\n        delta\nepsilon\n"""
        self.assertEqual( string_helper.fix_indents( text_a ), text_b )


class array_helper_tests( unittest.TestCase ):
    def test__create_index_lookup( self ):
        """
        TEST: create_index_lookup
        """
        
        my_list = "a", "b", "c", "d", "e"
        
        lookup = array_helper.create_index_lookup( my_list )
        
        # Sanity check
        for i, v in enumerate( my_list ):
            self.assertEqual( i, lookup[v] )


class batch_lists_test( unittest.TestCase ):
    def test__BatchList__take( self ):
        b = BatchList( (1, 2, 3, 4, 5, 6, 7, 8, 9, 10), 3 )
        
        self.assertEqual( b.take(), [1, 2, 3] )
        self.assertEqual( b.take(), [4, 5, 6] )
        self.assertEqual( b.take(), [7, 8, 9] )
        self.assertEqual( b.take(), [10, ] )
    
    
    def test__divide_workload( self ):
        self.assertEqual( batch_lists.divide_workload( 0, 5, 51 ), (0, 10) )
        self.assertEqual( batch_lists.divide_workload( 1, 5, 51 ), (10, 20) )
        self.assertEqual( batch_lists.divide_workload( 2, 5, 51 ), (20, 30) )
        self.assertEqual( batch_lists.divide_workload( 3, 5, 51 ), (30, 40) )
        self.assertEqual( batch_lists.divide_workload( 4, 5, 51 ), (40, 51) )
    
    
    def test__divide_workload_2( self ):
        last_b = None
        for i in range( 1000 ):
            a, b = batch_lists.divide_workload( i, 1000, 1600 )
            
            self.assertGreater( b, a )
            
            if last_b is not None:
                self.assertEqual( a, last_b )
            
            last_b = b


class reflection_helper_tests( unittest.TestCase ):
    def test__annotations( self ):
        class UPPER:
            pass
        
        
        class MIDDLE( UPPER ):
            pass
        
        
        class LOWER( MIDDLE ):
            pass
        
        
        class DIFFERENT:
            pass
        
        
        self.assertTrue( AnnotationInspector( MIDDLE ).is_indirectly_below( UPPER ) )
        self.assertTrue( AnnotationInspector( Optional[MIDDLE] ).is_indirectly_below( UPPER ) )
        self.assertTrue( AnnotationInspector( Optional[Union[DIFFERENT, MIDDLE]] ).is_indirectly_below( UPPER ) )
        self.assertFalse( AnnotationInspector( Optional[Union[DIFFERENT, MIDDLE]] ).is_indirectly_below( LOWER ) )
        
        self.assertTrue( AnnotationInspector( MIDDLE ).is_directly_below( UPPER ) )
        self.assertFalse( AnnotationInspector( Optional[MIDDLE] ).is_directly_below( UPPER ) )
        
        self.assertTrue( AnnotationInspector( MIDDLE ).is_indirectly_above( LOWER ) )
        self.assertTrue( AnnotationInspector( Optional[MIDDLE] ).is_indirectly_above( LOWER ) )
        self.assertTrue( AnnotationInspector( Optional[Union[DIFFERENT, MIDDLE]] ).is_indirectly_above( LOWER ) )
        self.assertFalse( AnnotationInspector( Optional[Union[DIFFERENT, MIDDLE]] ).is_indirectly_above( UPPER ) )
        
        self.assertTrue( AnnotationInspector( MIDDLE ).is_directly_above( LOWER ) )
        self.assertFalse( AnnotationInspector( Optional[MIDDLE] ).is_directly_above( LOWER ) )
        
        middle = MIDDLE()
        
        self.assertTrue( AnnotationInspector( MIDDLE ).is_viable_instance( middle ) )
        self.assertTrue( AnnotationInspector( Optional[UPPER] ).is_viable_instance( middle ) )
        self.assertTrue( AnnotationInspector( Union[LOWER, UPPER] ).is_viable_instance( middle ) )
        self.assertFalse( AnnotationInspector( Union[LOWER] ).is_viable_instance( middle ) )


if __name__ == '__main__':
    unittest.main()



class Colour:
    def __init__( self, r, g, b ):
        self.r = r
        self.g = g
        self.b = b
    
    @property
    def html(self):
        return "#{0:02x}{1:02x}{2:02x}".format( self.r, self.g, self.b )


BLACK = Colour( 0, 0, 0 )
DARK_GRAY = Colour( 64, 64, 64 )
GRAY = Colour( 128, 128, 128 )
LIGHT_GRAY = Colour( 192, 192, 192 )
WHITE = Colour( 255, 255, 255 )

RED = Colour( 255, 0, 0 )
GREEN = Colour( 0, 255, 0 )
BLUE = Colour( 0, 0, 255 )

CYAN = Colour( 0, 255, 255 )
MAGENTA = Colour( 255, 0, 255 )
YELLOW = Colour( 255, 255, 0 )

ORANGE = Colour( 255, 128, 0 )
FUSCHIA = Colour( 255, 0, 128 )
LIME = Colour( 128, 255, 0 )
SPRING_GREEN = Colour( 0, 255, 128 )
PURPLE = Colour( 128, 0, 255 )
DODGER_BLUE = Colour( 128, 0, 255 )

DARK_RED = Colour( 128, 0, 0 )
DARK_GREEN = Colour( 0, 128, 0 )
DARK_BLUE = Colour( 0, 0, 128 )

DARK_CYAN = Colour( 0, 128, 128 )
DARK_MAGENTA = Colour( 128, 0, 128 )
DARK_YELLOW = Colour( 128, 128, 0 )

DARK_ORANGE = Colour( 128, 64, 0 )
DARK_FUSCHIA = Colour( 128, 0, 64 )
DARK_LIME = Colour( 64, 128, 0 )
DARK_SPRINGGREEN = Colour( 0, 128, 64 )
DARK_PURPLE = Colour( 64, 0, 128 )
DARK_DODGERBLUE = Colour( 64, 0, 128 )

LIGHT_RED = Colour( 255, 128, 128 )
LIGHT_GREEN = Colour( 128, 255, 128 )
LIGHT_BLUE = Colour( 128, 128, 255 )

LIGHT_CYAN = Colour( 128, 255, 255 )
LIGHT_MAGENTA = Colour( 255, 0, 255 )
LIGHT_YELLOW = Colour( 255, 255, 128 )



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './designer/FrmText_designer.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Dialog(object):
    def __init__(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(678, 539)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.TXT_MAIN = QtWidgets.QTextEdit(Dialog)
        self.TXT_MAIN.setObjectName("TXT_MAIN")
        self.verticalLayout.addWidget(self.TXT_MAIN)
        self.BTNBOX_MAIN = QtWidgets.QDialogButtonBox(Dialog)
        self.BTNBOX_MAIN.setOrientation(QtCore.Qt.Horizontal)
        self.BTNBOX_MAIN.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.BTNBOX_MAIN.setObjectName("BTNBOX_MAIN")
        self.verticalLayout.addWidget(self.BTNBOX_MAIN)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))




"""
Things for dealing with filenames
"""
import warnings
from os import path
import os

import errno
from typing import Iterable, List, Optional


def read_all_text( file_name: str ) -> Optional[str]:
    """
    Reads all the text from a file, returning None if there is no file
    """
    
    if not os.path.isfile( file_name ):
        return None
    
    with open( file_name, "r" ) as file:
        return file.read()


def get_subdirs( directory_name: str ) -> List[str]:
    results = []
    
    for file_name in os.listdir( directory_name ):
        full_name = os.path.join( directory_name, file_name )
        if path.isdir( full_name ):
            results.append( full_name )
    
    return results


def read_all_lines( file_name: str ) -> Optional[List[str]]:
    """
    Reads all the text from a file, returning None if there is no file
    """
    
    if not os.path.isfile( file_name ):
        return None
    
    with open( file_name, "r" ) as file:
        return list( x.rstrip( "\n" ) for x in file )


def write_all_text( file_name: object, text: str ) -> None:
    """
    Writes all text to a file, overwriting the existing content
    """
    
    with open( str( file_name ), "w" ) as file:
        file.write( text )


def contains_files( directory: str, ext: str ) -> bool:
    """
    Returns if the directory contains any files.
    """
    ext = ext.upper()
    
    for root, dirs, files in os.walk( directory ):
        for file in files:
            if file.upper().endswith( ext ):
                return True
    
    return False


def get_file_name( full_path: str ) -> str:
    """
    Returns <FILE><EXT> from <PATH><FILE><EXT>
    `a/b/c.d` --> `c.d`
    """
    return path.split( full_path )[1]


def replace_extension( file_name: str, new_extension: str ) -> str:
    """
    Replaces <EXT> in <PATH><FILE><EXT>
    `a/b/c.d` <-- `d`
    """
    return path.splitext( file_name )[0] + new_extension


def get_extension( file_name: str ) -> str:
    """
    Returns <EXT> in <PATH><FILE><EXT>. Note this includes the ".".
    `a/b/c.d` --> `.d`
    """
    return path.splitext( file_name )[1]


def get_filename( file_name: str ) -> str:
    """
    Returns <FILE><EXT> in <PATH><FILE><EXT>
    `a/b/c.d` --> `c.d`
    """
    return path.split( file_name )[1]


def get_filename_without_extension( file_name: str ) -> str:
    """
    Returns <FILE> from <PATH><FILE><EXT>
    `a/b/c.d` --> `c`
    """
    file_name = path.split( file_name )[1]
    file_name = path.splitext( file_name )[0]
    return file_name


def get_full_filename_without_extension( file_name: str ) -> str:
    """
    Returns <PATH><FILE> from <PATH><FILE><EXT>
    `a/b/c.d` --> `a/b/c`
    """
    return path.splitext( file_name )[0]


def replace_filename_without_extension( file_name: str, new_name: str ) -> str:
    """
    Replaces <NAME> in <PATH><NAME><EXT>
    `a/b/c.d` <-- `c`
    """
    return path.join( get_directory( file_name ), new_name + get_extension( file_name ) )


def replace_filename( file_name: str, new_name: str ):
    """
    Replaces <NAME><EXT> in <PATH><NAME><EXT>
    `a/b/c.d` --> `c.d`
    """
    return path.join( get_directory( file_name ), new_name )


def join( *args, **kwargs ):
    """
    Just calls path.join
    """
    return path.join( *args, **kwargs )


def get_directory( file_name: str, up = 1 ) -> str:
    """
    Returns <PATH> in <PATH><NAME><EXT>
    """
    for _ in range( up ):
        file_name = path.split( file_name )[0]
    
    return file_name


def suffix_directory( file_name: str ) -> str:
    if file_name.endswith( path.sep ):
        return file_name
    else:
        return file_name + path.sep


def relocate( target_files: Iterable[str], new_folder: str, locate: bool ) -> None:
    """
    Given a list of files, creates a set of symbolic links to them in another folder. 
    :param target_files: List of files (complete paths unless locate is False, in which case they can be partial paths) 
    :param new_folder: Folder to create links in or remove links from  
    :param locate: Whether to create links (True) or remove them (False) 
    """
    for file in target_files:
        name = get_file_name( file )
        new_name = path.join( new_folder, name )
        
        if locate:
            os.link( file, new_name )
        else:
            os.unlink( new_name )


def split_path( path_: str ) -> List[str]:
    """
    Splits a name into its folders and file
    """
    
    folders = []
    
    while 1:
        path_, folder = os.path.split( path_ )
        
        if folder != "":
            folders.append( folder )
        else:
            if path_ != "":
                folders.append( path_ )
            break
    
    folders.reverse()
    
    return folders


def sub_dirs( directory ):
    result = list( path.join( directory, x ) for x in os.listdir( directory ) )
    
    result = [x for x in result if path.isdir( x )]
    
    return result


def list_dir( directory, filter = None, recurse = False ) -> List[str]:
    if recurse:
        result = []
        for folder, subfolders, files in os.walk( directory ):
            for file in files:
                result.append( path.join( folder, file ) )
    else:
        result = list( path.join( directory, x ) for x in os.listdir( directory ) )
    
    if filter:
        filter = filter.upper()
        result = [x for x in result if x.upper().endswith( filter )]
    
    return result


def is_windows():
    return os.name == "nt"


def default_extension( file_name, extension ):
    if not get_extension( file_name ):
        return file_name + extension
    
    return file_name


def create_directory( output_directory: str ):
    """
    Creates a directory (doesn't do anything if it already exists)
    """
    
    if path.isdir( output_directory ):
        return
    
    try:
        os.makedirs( output_directory )
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise FileNotFoundError( "Failed to create directory «{}» due to another error.".format( output_directory ) ) from exception
    
    if not path.isdir( output_directory ):
        raise FileNotFoundError( "Command returned success but the created directory cannot be found «{}».".format( output_directory ) )


def file_len( file_name: str, line_start: Optional[str] = None ) -> int:
    """
    Counts the number of lines in a file, optionally those just starting "line_start"
    """
    
    c = 0
    
    with open( file_name ) as f:
        for line in f:
            if (not line_start) or (line.startswith( line_start )):
                c += 1
    
    return c


def safe_file_name( name ):
    name = str( name )
    name = name.replace( "<>:\"'/\\|?*", "_" )
    
    if not name:
        return "Untitled"
    else:
        return name


def recycle_file( file_name ):
    try:
        from send2trash import send2trash
        send2trash( file_name )
    except ImportError:
        # noinspection PyUnusedLocal
        send2trash = None
        warnings.warn( "send2trash", ImportWarning )
        os.remove( file_name )


def get_last_directory_and_filename( file_name ):
    """
    From `a/b/c/d.e` gets `c/d.e` 
    """
    return path.join( get_filename( get_directory( file_name ) ), get_filename( file_name ) )


def delete_file( file_name: str ):
    if path.isfile( file_name ):
        os.remove( file_name )


def home():
    return path.expanduser( "~" )  # TODO: Does this work on Windows?


def file_size( file_name: str ) -> int:
    """
    Size of a file, or -1 on error.
    """
    try:
        return os.stat( file_name ).st_size
    except:
        return -1


def sequential_file_name( file_name: str ) -> str:
    """
    DEPRECATED
    Please use CacheHelper instead.
    
    Creates an output file in the $(APP_DISPLAY_NAME) temporary folder.
    Avoids conflicts with existing filenames.
    NO MULTI-THREAD SUPPORT - Only acknowledges existing files.
    
    :param file_name: Format of filename, with * where the number goes.
    :return: Full filename
    """
    if "*" not in file_name:
        raise ValueError( "`sequential_file_name` requires the filename to contain a placeholder '*' to represent the number, but the value provided «{}» does not.".format( file_name ) )
    
    number = 1
    
    result = file_name.replace( "*", str( number ) )
    
    while path.exists( result ):
        number += 1
        result = file_name.replace( "*", str( number ) )
    
    return result


def highlight_file_name_without_extension( file, highlight, normal ):
    return normal + path.join( get_directory( file ), highlight + get_filename_without_extension( file ) + normal + get_extension( file ) )



# The following are decorators, they don't do anything, they're just used as comments
from inspect import isclass


def ignore( *_, **__ ):
    """
    Ignores its parameters
    """
    pass

def protected( f ):
    return f


def sealed( f ):
    return f


def override( f ):  # means "I'm not documenting this because the documentation is in the base class"
    return f


def overrides( interface ):
    ignore(interface)
    return override


def virtual( f ):
    return f


def abstract( f ):
    if isclass( f ):
        return f
    
    
    def fn( self, *_, **__ ):
        raise NotImplementedError( "An attempt has been made to call a method but this object has an abstract method «{}». The type of object is «{}» and its string representation is «{}».".format( f.__name__, type( self ).__name__, repr( self ) ) )
    
    
    return fn






import unittest


class MathsHelper_test( unittest.TestCase ):
    def test_increment_mean( self ):
        from mhelper import maths_helper
        
        my_values = [ 1, 7, 23, 2, 6, 2, 6, 8, 42, 9, 3, 1, 0.2, 0.3 ]
        
        incremental_average = 0
        incremental_count = 0
        
        for index, new_value in enumerate( my_values ):
            incremental_average, incremental_count = maths_helper.increment_mean( incremental_average, incremental_count, new_value )
            true_average = sum( my_values[ :index + 1 ] ) / (index + 1)
            self.assertAlmostEqual( incremental_average, true_average )
            self.assertEqual( incremental_count, index + 1 )


if __name__ == '__main__':
    unittest.main()



class StringParser:
    def __init__( self, text: str ):
        self.text = text
        self.cursor = 0
    
    
    def read_char( self ):
        return self.read_text( 1 )
    
    
    def read_text( self, length: int ):
        result = self.text[ self.cursor:self.cursor + length ]
        self.cursor += length
        return result
    
    
    def read_to( self, find: str ):
        start = self.cursor
        end = self.text.find( find, start )
        
        if end == -1:
            end = len( self.text )
            
        self.cursor = end
        
        return self.text[ start:end ]
    
    
    def read_past( self, find: str ):
        result = self.read_to( find )
        self.cursor += len( find )
        return result
    
    
    def read_all( self ):
        return self.read_text( len( self.text ) - self.cursor )
    
    
    def end( self ):
        return self.cursor >= len( self.text )


def string_parser_unit_test():
    TEST_STRING = "Once I had a lemon."
    
    parser = StringParser( TEST_STRING )
    
    # |O|nce I had a lemon.
    x = parser.read_char()
    print(x)
    assert x == "O", "[{}]".format(x)
    
    # O|nce |I had a lemon.
    x = parser.read_text( 4 )
    print(x)
    assert x == "nce ", "[{}]".format(x)
    
    # Once |I| had a lemon.
    x = parser.read_to( " " )
    print(x)
    assert x == "I", "[{}]".format(x)
    
    # Once I| |had a lemon.
    x = parser.read_char()
    print(x)
    assert x == " ", "[{}]".format(x)
    
    # Once I |had| |a lemon.
    x = parser.read_past( " " )
    print(x)
    assert x == "had"
    
    # Once I had |a| lemon.
    x = parser.read_char()
    print(x)
    assert x == "a"
    
    assert not parser.end()
    
    # Once I had a| lemon.|
    x = parser.read_all()
    print(x)
    assert x == " lemon.", "[{}]".format(x)
    
    assert parser.end()



from distutils.core import setup


setup( name = "mhelper",
       url = "https://bitbucket.org/mjr129/mhelper",
       version = "1.0.1.23",
       description = "Includes a collection of utility functions.",
       author = "Martin Rusilowicz",
       license = "GNU AGPLv3",
       packages = ["mhelper",
                   "mhelper.qt_helpers"],
       install_requires = ["PyQt5"]
       )



import traceback


def print_exception(exception, traceback_ = None):
    if not traceback_:
        traceback_ = traceback.format_exc()
        
    print( str( exception ) )
    print( traceback_ )




import subprocess
import traceback


def add_details( exception: Exception, **kwargs ) -> None:
    args = list( exception.args )
    
    message = create_details_message( **kwargs )
    
    if len( args ) > 0 and isinstance( args[0], str ):
        args[0] += message
    else:
        args.append( message )
    
    exception.args = tuple( args )


def create_details_message( **kwargs ):
    from mhelper import string_helper
    
    result = [""]
    
    lk = 1
    lt = 1
    
    for k, v in kwargs.items():
        lk = max( len( str( k ) ), lk )
        lt = max( len( string_helper.type_name( v ) ), lt )
    
    for k, v in kwargs.items():
        result.append( "--> {0} ({1}) = «{2}»".format( str( k ).ljust( lk ), string_helper.type_name( v ).ljust( lt ), v ) )
    
    return "\n".join( result )


def assert_type( name, value, type ):
    if not isinstance( value, type ):
        from mhelper.string_helper import type_name
        raise TypeError( "`{0}` should be of type `{1}`, but it is a `{2}` with value `{3}`.".format( name, type.__name__, type_name( value ), value ) )


def exception_to_string( ex: BaseException ):
    result = []
    
    while ex:
        result.append( str( ex ) )
        ex = ex.__cause__
    
    return "\n---CAUSED BY---\n".join( result )


class NotSupportedError( Exception ):
    """
    Since `NotImplementedError` looks like an abstract-base-class error to the IDE you can use `NotSupportedError` if desired.
    """
    pass


class ImplementationError( Exception ):
    pass


class LogicError( Exception ):
    pass


class NotFoundError( Exception ):
    pass


class SwitchError( Exception ):
    def __init__( self, name, value, *, instance = False, details = None ):
        if details is not None:
            details = " Further details: {}".format( details )
        
        if instance:
            super().__init__( "The switch on the type of «{}» does not recognise the value «{}» of type «{}».{}".format( name, value, type( value ), details ) )
        else:
            super().__init__( "The switch on «{}» does not recognise the value «{}» of type «{}».{}".format( name, value, type( value ), details ) )


class SubprocessError( Exception ):
    pass


def run_subprocess( command ):
    status = subprocess.call( command, shell = True )
    
    if status:
        raise SubprocessError( "The command «{}» exited with error code {}. If available, checking the console output may provide more details.".format( command, status ) )


def assert_instance( name: str, value: object, type_: type ):
    if not isinstance( value, type_ ):
        raise TypeError( instance_message( name, value, type_ ) )


def assert_instance_or_none( name: str, value: object, type_: type ):
    if value is not None:
        if not isinstance( value, type_ ):
            raise TypeError( "The value of «{}», which is «{}», should either be `None` or an object of type «{}», but it's not, it's a «{}».".format( name, value, type_, type( value ) ) )


def instance_message( name: str, value: object, type_: type ):
    return "The value of «{}», which is «{}», should be of type «{}», but it's not, it's a «{}».".format( name, value, type_, type( value ) )


def full_traceback():
    return "Handler Traceback:\n" + current_stack_text() + "\nError " + traceback.format_exc()


def current_stack_text():
    return "\n".join( x.strip() for x in traceback.format_stack() )



from importlib import import_module
from types import ModuleType
from typing import List, Union

from os import path

from mhelper import file_helper


def load_namespace( module_: Union[ ModuleType, str ] ) -> None:
    """
    Registers a all plugins within a module or namespace.
    """
    if isinstance( module_, str ):
        module_ = import_module( module_ )
    
    for sub_module_name in list_namespace( module_ ):
        load_namespace( sub_module_name )


def list_namespace( module: ModuleType ) -> List[ str ]:
    """
    Given a module (i.e. a simple folder) finds the names of all sub-modules.
    
    :param module:    Module to search. 
    :return:          A list of module names, or if the module isn't a folder, an empty list. 
    """
    folder = get_module_path( module )
    
    if not path.isdir( folder ):
        return [ ]
    
    results = [ ]
    
    for file_name in file_helper.list_dir( folder, ".py" ) + file_helper.sub_dirs( folder ):
        module_name = file_helper.get_filename_without_extension( file_name )
        
        if not module_name.startswith( "_" ):
            module_name = module.__name__ + "." + module_name
            results.append( module_name )
    
    return results


def get_module_path( module: ModuleType ) -> str:
    """
    Gets the path to a module.
    """
    return module.__file__ if hasattr( module, "__file__" ) else module.__path__._path[ 0 ]



__author__ = "Martin Rusilowicz"



class ResourceIcon:
    def __init__( self, path: str ):
        self._path = path
        self._icon = None
        
    def __repr__(self):
        return "ResourceIcon(«{0}»)".format(self._path)


    def icon( self ):
        from PyQt5.QtGui import QIcon
        
        if not self._icon:
            self._icon = QIcon( self._path )

        return self._icon
